"""
Virt-v2v test utility functions.

:copyright: 2008-2012 Red Hat Inc.
"""

from __future__ import print_function

import os
import re
import time
import glob
import logging
import random
import uuid
import aexpect
import shutil

from aexpect import remote

from avocado.utils import path
from avocado.utils import process
from avocado.utils.astring import to_text
from avocado.core import exceptions

from virttest import libvirt_vm as lvirt
from virttest import ovirt
from virttest import virsh
from virttest import ppm_utils
from virttest import data_dir
from virttest import remote as remote_old
from virttest import utils_misc
from virttest import ssh_key
from virttest.utils_test import libvirt
from virttest.libvirt_xml import vm_xml
from virttest.utils_version import VersionInterval
from virttest.utils_misc import asterisk_passwd
from virttest.utils_misc import compare_md5

try:
    V2V_EXEC = path.find_command('virt-v2v')
except path.CmdNotFoundError:
    V2V_EXEC = None


class Uri(object):

    """
    This class is used for generating uri.
    """

    def __init__(self, hypervisor):
        if hypervisor is None:
            # kvm is a default hypervisor
            hypervisor = "kvm"
        self.hyper = hypervisor

    def get_uri(self, hostname, vpx_dc=None, esx_ip=None):
        """
        Uri dispatcher.

        :param hostname: String with host name.
        """
        uri_func = getattr(self, "_get_%s_uri" % self.hyper)
        self.host = hostname
        self.vpx_dc = vpx_dc
        self.esx_ip = esx_ip
        return uri_func()

    def _get_kvm_uri(self):
        """
        Return kvm uri.
        """
        uri = "qemu:///system"
        return uri

    def _get_xen_uri(self):
        """
        Return xen uri.
        """
        uri = "xen+ssh://root@" + self.host + "/"
        return uri

    def _get_esx_uri(self):
        """
        Return esx uri.
        """
        uri = ''
        if self.vpx_dc and self.esx_ip:
            uri = "vpx://root@%s/%s/%s/?no_verify=1" % (self.host,
                                                        self.vpx_dc,
                                                        self.esx_ip)
        if not self.vpx_dc and self.esx_ip:
            uri = "esx://root@%s/?no_verify=1" % self.esx_ip
        return uri

    # add new hypervisor in here.


class Target(object):

    """
    This class is used for generating command options.
    """

    def __init__(self, target, uri):
        if target is None:
            # libvirt is a default target
            target = "libvirt"
        self.target = target
        self.uri = uri
        # Save NFS mount records like {0:(src, dst, fstype)}
        self.mount_records = {}
        # Authorized_keys is a list which every element is a tuple.
        # The value of each tuple is like (session, key, server_type).
        self.authorized_keys = []

    def cleanup(self):
        """
        Cleanup NFS mount records
        """
        for src, dst, fstype in self.mount_records.values():
            utils_misc.umount(src, dst, fstype)

        self.cleanup_authorized_keys()

    def cleanup_authorized_keys(self):
        """
        Clean up authorized_keys in remote server
        """
        try:
            for session, key, server_type in self.authorized_keys:
                if not session or not key:
                    continue
                logging.debug(
                    "session=%s key=%s server_type=%s",
                    session,
                    key,
                    server_type)
                session.cmd("sed -i '/%s/d' %s" %
                            (key, get_authorized_keys_file(server_type)))
        finally:
            # Close session here in case the following element use same
            # session, although it could not happen in general.
            for session, _, _ in self.authorized_keys:
                if session:
                    logging.debug("closed session = %s", session)
                    session.close()

    def get_cmd_options(self, params):
        """
        Target dispatcher.
        """
        self.params = params
        self.output_method = self.params.get('output_method', 'rhv')
        self.input_mode = self.params.get('input_mode')
        self.esxi_host = self.params.get(
            'esxi_host', self.params.get('esx_ip'))
        self.datastore = self.params.get('datastore')
        self._nfspath = self.params.get('_nfspath')
        self.src_uri_type = params.get('src_uri_type')
        self.esxi_password = params.get('esxi_password')
        self.input_transport = self.params.get('input_transport')
        self.vddk_libdir = self.params.get('vddk_libdir')
        self.vddk_libdir_src = self.params.get('vddk_libdir_src')
        self.vddk_thumbprint = self.params.get('vddk_thumbprint')
        self.vcenter_host = self.params.get('vcenter_host')
        self.vcenter_password = self.params.get('vcenter_password')
        self.vm_name = self.params.get('main_vm')
        self.bridge = self.params.get('bridge')
        self.network = self.params.get('network')
        self.of_format = self.params.get('of_format', 'raw')
        self.input_file = self.params.get('input_file')
        self.new_name = self.params.get('new_name')
        self.username = self.params.get('username', 'root')
        self.vmx_nfs_src = self.params.get('vmx_nfs_src')
        self.has_genid = self.params.get('has_genid')
        # --mac arguments with format as v2v, multiple macs can be
        # sperated by ';'.
        self.iface_macs = self.params.get('iface_macs')
        # '_iface_list' is set automatically, Users should not use it.
        self._iface_list = self.params.get('_iface_list')
        self.net_vm_opts = ""
        self._vmx_filename_fullpath = self._vmx_filename = ""

        def _compose_vmx_filename():
            """
            Return vmx filename for '-i vmx'.

            All vmname, nfs directory and vmx name may be different
            e.g.
            vmname:  esx6.7-ubuntu18.04-x86_64
            nfspath: esx6.5-ubuntu18.04-x86_64-bug1481930/esx6.5-ubuntu18.04-x86_64-bug1481930.vmx
            """
            if self.vmx_nfs_src and ':' in self.vmx_nfs_src:
                mount_point = v2v_mount(self.vmx_nfs_src, 'vmx_nfs_src')
                self.mount_records[len(self.mount_records)] = (
                    self.vmx_nfs_src, mount_point, None)

                vmx_path = os.path.join(mount_point, self._nfspath, '*.vmx')
                vmxfiles = glob.glob(vmx_path)

                if len(vmxfiles) == 0:
                    raise exceptions.TestError(
                        "Did not found any vmx files in %s" % vmx_path)

                self._vmx_filename_fullpath = vmxfiles[0]
                self._vmx_filename = os.path.basename(vmxfiles[0])
                logging.debug(
                    'vmx file full path is %s' %
                    self._vmx_filename_fullpath)
            else:
                # This only works for -i vmx -it ssh, because it only needs an vmx filename,
                # and doesn't have to mount the nfs storage. If the guessed name is wrong,
                # v2v will report an error.
                logging.info(
                    'vmx_nfs_src is not set in cfg file, try to guess vmx filename')
                # some guest's direcotory name ends with '_1',
                # e.g. esx5.5-win10-x86_64_1/esx5.5-win10-x86_64.vmx
                #
                # Note: the pattern order cannot be changed
                guess_ptn_list = [r'(^.*?(x86_64|i386))_[0-9]+$',
                                  r'(^.*?(x86_64|i386)$)',
                                  r'(^.*?)_[0-9]+$']

                for ptn in guess_ptn_list:
                    if re.search(ptn, self._nfspath):
                        self._vmx_filename = re.search(
                            ptn, self._nfspath).group(1)
                        break

                if not self._vmx_filename:
                    self._vmx_filename = self._nfspath

                self._vmx_filename = self._vmx_filename + '.vmx'
                logging.debug(
                    'Guessed vmx file name is %s' %
                    self._vmx_filename)

        def _compose_input_transport_options():
            """
            Set input transport options for v2v
            """
            options = ''
            if self.input_transport is None:
                return options

            # -it vddk
            if self.input_transport == 'vddk':
                if self.vddk_libdir is None or not os.path.isdir(
                        self.vddk_libdir):
                    # Invalid nfs mount source if no ':'
                    if self.vddk_libdir_src is None or ':' not in self.vddk_libdir_src:
                        logging.error(
                            'Neither vddk_libdir nor vddk_libdir_src was set')
                        raise exceptions.TestError(
                            "VDDK library directory or NFS mount point must be set")

                    vddk_lib_prefix = 'vddklib_'
                    # General vddk directory
                    vddk_lib_rootdir = '/var/tmp/vddk_libdir'
                    vddk_libdir = '%s/latest' % vddk_lib_rootdir
                    check_list = ['FILES',
                                  'lib64/libgvmomi.so',
                                  'bin64/vmware-vdiskmanager']

                    mount_point = v2v_mount(
                        self.vddk_libdir_src, 'vddk_libdir')

                    logging.info('Preparing vddklib on local server')
                    if os.path.exists(vddk_lib_rootdir):
                        if os.path.exists(vddk_libdir):
                            os.unlink(vddk_libdir)

                        vddklib_count = len(
                            glob.glob(
                                '%s/%s*' %
                                (vddk_lib_rootdir, vddk_lib_prefix)))
                        for i in range(1, vddklib_count + 1):
                            vddk_lib = '%s/%s' % (vddk_lib_rootdir,
                                                  vddk_lib_prefix + str(i))
                            if all(compare_md5(os.path.join(mount_point, file_i), os.path.join(
                                    vddk_lib, file_i)) for file_i in check_list):
                                os.symlink(vddk_lib, vddk_libdir, True)
                                break

                        if not os.path.exists(vddk_libdir):
                            vddk_lib = '%s/%s' % (vddk_lib_rootdir,
                                                  vddk_lib_prefix + str(vddklib_count + 1))
                            shutil.copytree(mount_point, vddk_lib)
                            os.symlink(vddk_lib, vddk_libdir, True)
                    else:
                        vddk_lib = '%s/%s' % (vddk_lib_rootdir,
                                              vddk_lib_prefix + '1')
                        shutil.copytree(mount_point, vddk_lib)
                        os.symlink(vddk_lib, vddk_libdir, True)

                    logging.info('vddklib on local server is %s', vddk_lib)
                    self.vddk_libdir = vddk_libdir
                    utils_misc.umount(self.vddk_libdir_src, mount_point, None)

                # Invalid vddk thumbprint if no ':'
                if self.vddk_thumbprint is None or ':' not in self.vddk_thumbprint:
                    self.vddk_thumbprint = get_vddk_thumbprint(
                        *
                        (
                            self.esxi_host,
                            self.esxi_password,
                            self.src_uri_type) if self.src_uri_type == 'esx' else (
                            self.vcenter_host,
                            self.vcenter_password,
                            self.src_uri_type))

            # -it ssh
            if self.input_transport == 'ssh':
                pub_key, session = v2v_setup_ssh_key(
                    self.esxi_host, self.username, self.esxi_password, server_type='esx', auto_close=False)
                self.authorized_keys.append(
                    (session, pub_key.split()[1].split('/')[0], 'esx'))
                utils_misc.add_identities_into_ssh_agent()

            # New input_transport type can be added here

            # A dict to save input_transport types and their io options, new it_types
            # should be added here. Their io values were composed during running time
            # based on user's input
            input_transport_args = {
                'vddk': "-io vddk-libdir=%s -io vddk-thumbprint=%s" % (self.vddk_libdir,
                                                                       self.vddk_thumbprint),
                'ssh': "ssh://root@{}/vmfs/volumes/{}/{}/{}".format(
                    self.esxi_host,
                    self.datastore,
                    self._nfspath,
                    self._vmx_filename)}

            options = " -it %s " % (self.input_transport)
            options += input_transport_args[self.input_transport]
            return options

        supported_mac = v2v_supported_option(
            r'--mac <mac:network\|bridge(\|ip)?:out>')
        if supported_mac:
            if self.iface_macs:
                for mac_i in self.iface_macs.split(';'):
                    # [mac, net_type, net], e.x. ['xx:xx:xx:xx:xx:xx', 'bridge', 'virbr0']
                    mac_i_list = mac_i.rsplit(':', 2)
                    # Just warning invalid values in case for negative testing
                    if len(mac_i_list) != 3 or mac_i_list[1] not in [
                            'bridge', 'network']:
                        logging.warning(
                            "Invalid value for --mac '%s'" %
                            mac_i_list)
                    mac, net_type, netname = mac_i_list
                    self.net_vm_opts += " --mac %s:%s:%s" % (
                        mac, net_type, netname)
            else:
                logging.info("auto set --mac option")
                for mac, _ in self._iface_list:
                    # Randomly cover both 'bridge' and 'network' even thought there is no
                    # difference.
                    if random.choice(['bridge', 'network']) == 'network':
                        self.net_vm_opts += " --mac %s:%s:%s" % (
                            mac, 'network', self.network)
                    else:
                        self.net_vm_opts += " --mac %s:%s:%s" % (
                            mac, 'bridge', self.bridge)

        if self.input_mode == 'vmx':
            _compose_vmx_filename()

        if not self.net_vm_opts:
            if supported_mac:
                logging.warning("auto set --mac failed, roll back to -b/-n")
            if self.bridge:
                self.net_vm_opts += " -b %s" % self.bridge
            if self.network:
                self.net_vm_opts += " -n %s" % self.network

        if self.input_mode != 'vmx':
            self.net_vm_opts += " %s" % self.vm_name
        elif self.input_transport is None:
            self.net_vm_opts += " %s" % self._vmx_filename_fullpath

        options = self.get_target_options() + _compose_input_transport_options()

        if self.new_name:
            options += ' -on %s' % self.new_name

        if self.input_mode is not None:
            options = " -i %s %s" % (self.input_mode, options)
            if self.input_mode in ['ova', 'disk',
                                   'libvirtxml'] and self.input_file:
                options = options.replace(self.vm_name, self.input_file)

        # In '-i vmx', '-ic' is not needed
        if self.input_mode == 'vmx':
            options = re.sub(r'-ic .*? ', '', options)
        return options

    def _get_os_directory(self):
        """
        Prepare a local os directory for v2v conversion.

        Correspond to '-os DIRECTORY'.
        """
        os_directory = self.params.get('os_directory')

        if not os_directory:
            os_directory = os.path.join(
                data_dir.get_tmp_dir(), 'v2v_os_directory')
            if not os.path.exists(os_directory):
                os.makedirs(os_directory)
        # Pass the json directory to testcase for checking
        self.params.get('params').update({'os_directory': os_directory})

        logging.debug('The os directory(-os DIRECTORY) is %s.', os_directory)
        return os_directory

    def _get_libvirt_options(self):
        """
        Construct output options for -o libvirt

        'os_pool' corresponds to '-os POOL'.
        """
        os_pool = self.params.get('os_pool')
        options = " -os %s" % os_pool

        return options

    def _get_ovirt_options(self):
        """
        Construct output options for -o ovirt

        'os_storage' corresponds to '-o rhv -os [esd:/path|/path'.
        'os_storage_name' corresponds to '-o rhv -os STORAGE'.
        'rhv_upload_opts' includes all '-oo xxx' options.
        """
        os_storage = self.params.get('os_storage')
        os_storage_name = self.params.get('os_storage_name')
        rhv_upload_opts = self.params.get('rhv_upload_opts')
        output_method = self.output_method
        options = " -os %s" % (os_storage if output_method !=
                               "rhv_upload" else os_storage_name)
        if rhv_upload_opts:
            options += ' %s' % rhv_upload_opts

        return options

    def _get_local_options(self):
        """
        Construct output options for -o local
        """
        os_directory = self._get_os_directory()
        options = " -os %s" % os_directory

        return options

    def _get_null_options(self):
        """
        Construct output options for -o null
        """
        return ''

    def _get_json_options(self):
        """
        Construct output options for -o json

        'oo_json_disk_pattern' corresponds to '-o json [-oo json-disks-pattern=PATTERN]'
        """
        oo_json_disk_pattern = self.params.get('oo_json_disk_pattern')
        os_directory = self._get_os_directory()

        options = " -os %s" % os_directory
        if oo_json_disk_pattern:
            options += " -oo json-disks-pattern=%s" % oo_json_disk_pattern

        return options

    def get_target_options(self):
        """
        Return command options.
        """
        uri = self.uri
        target = self.target
        o_fmt = self.of_format
        _get_target_specific_options = getattr(
            self, "_get_%s_options" % self.target)

        if target == 'ovirt':
            target = self.output_method.replace('_', '-')

        options = " -ic %s -o %s -of %s" % (uri, target, o_fmt)
        options += _get_target_specific_options() + self.net_vm_opts

        return options

    # add new target in here.


class VMCheck(object):

    """
    This is VM check class dispatcher.
    """

    def __new__(cls, test, params, env):
        # 'linux' is default os type
        os_type = params.get('os_type', 'linux')

        if cls is VMCheck:
            class_name = eval(os_type.capitalize() + str(cls.__name__))
            return super(VMCheck, cls).__new__(class_name)
        else:
            return super(VMCheck, cls).__new__(cls, test, params, env)

    def __init__(self, test, params, env):
        self.vm = None
        self.test = test
        self.env = env
        self.params = params
        self.name = params.get('main_vm')
        self.os_version = params.get("os_version")
        self.os_type = params.get('os_type', 'linux')
        self.target = params.get('target')
        self.username = params.get('vm_user', 'root')
        self.password = params.get('vm_pwd')
        self.nic_index = params.get('nic_index', 0)
        self.export_name = params.get('export_name')
        self.delete_vm = 'yes' == params.get('vm_cleanup', 'yes')
        self.virsh_session = params.get('virsh_session')
        self.virsh_session_id = self.virsh_session.get_id(
            ) if self.virsh_session else params.get('virsh_session_id')
        self.windows_root = params.get("windows_root", r"C:\WINDOWS")
        self.output_method = params.get("output_method")
        # Need create session after create the instance
        self.session = None

        if self.name is None:
            logging.error("vm name not exist")

        # libvirt is a default target
        if self.target == "libvirt" or self.target is None:
            self.vm = lvirt.VM(self.name, self.params, self.test.bindir,
                               self.env.get("address_cache"))
            self.pv = libvirt.PoolVolumeTest(test, params)
        elif self.target == "ovirt":
            self.vm = ovirt.VMManager(self.name, self.params, self.test.bindir,
                                      self.env.get("address_cache"))
        else:
            raise ValueError("Doesn't support %s target now" % self.target)

    def create_session(self, timeout=480):
        if self.session:
            logging.debug('vm session %s exists', self.session)
            return
        self.session = self.vm.wait_for_login(nic_index=self.nic_index,
                                              timeout=timeout,
                                              username=self.username,
                                              password=self.password)
        logging.debug('A new vm session %s was created', self.session)

    def cleanup(self):
        """
        Cleanup VM and remove all of storage files about guest
        """
        if self.session:
            logging.debug('vm session %s is closing', self.session)
            self.session.close()
            self.session = None

        # If VMChecker is instantiated before import_vm_to_ovirt and
        # the VMChecker.run is skiped, self.vm.instance will be NULL.
        # The update_instance should be ran before cleaning up.
        if hasattr(self.vm, 'update_instance'):
            self.vm.update_instance()
        if self.vm.instance and self.vm.is_alive():
            self.vm.destroy(gracefully=False)
            time.sleep(5)

        if self.target == "libvirt":
            if self.vm.exists() and self.vm.is_persistent():
                self.vm.undefine()

        if self.target == "ovirt":
            logging.debug("Deleting VM %s in Ovirt", self.name)
            self.vm.delete()
            # When vm is deleted, the disk will also be removed from
            # data domain, so it's not necessary to delete disk from
            # export domain for rhv_upload.
            if self.output_method != "rhv_upload":
                self.vm.delete_from_export_domain(self.export_name)
            ovirt.disconnect()

    def storage_cleanup(self):
        """
        Cleanup storage pool and volume
        """
        raise NotImplementedError

    def run_cmd(self, cmd, debug=True):
        """
        Run command in VM

        If cmd is a list or tuple, the run_cmd tries every element of
        cmd until success.
        """

        if isinstance(cmd, str):
            status, output = self.session.cmd_status_output(cmd)
        elif isinstance(cmd, list) or isinstance(cmd, tuple):
            for cmd_i in cmd:
                status, output = self.session.cmd_status_output(cmd_i)
                if status == 0:
                    break
        else:
            raise exceptions.TestError("Incorrect cmd: %s" % cmd)

        if debug:
            logging.debug("Command return status: %s", status)
            logging.debug("Command output:\n%s", output)
        return status, output


class LinuxVMCheck(VMCheck):

    """
    This class handles all basic linux VM check operations.
    """

    def get_vm_kernel(self):
        """
        Get vm kernel info.
        """
        cmd = "uname -r"
        return self.run_cmd(cmd)[1]

    def get_vm_os_info(self):
        """
        Get vm os info.
        """
        os_info = ""
        try:
            cmd = "cat /etc/os-release"
            status, output = self.run_cmd(cmd)
            if status != 0:
                cmd = "cat /etc/issue"
                output = self.run_cmd(cmd)[1]
                os_info = output.splitlines()[0]
            else:
                os_info = re.search(r'PRETTY_NAME="(.+)"', output).group(1)
        except Exception as e:
            logging.error("Fail to get os distribution: %s", e)
        return os_info

    def get_vm_os_vendor(self):
        """
        Get vm os vendor.
        """
        os_info = self.get_vm_os_info()
        if re.search('Red Hat', os_info):
            vendor = 'Red Hat'
        elif re.search('Fedora', os_info):
            vendor = 'Fedora Core'
        elif re.search('SUSE', os_info):
            vendor = 'SUSE'
        elif re.search('Ubuntu', os_info):
            vendor = 'Ubuntu'
        elif re.search('Debian', os_info):
            vendor = 'Debian'
        else:
            vendor = 'Unknown'
        logging.debug("The os vendor of VM '%s' is: %s" %
                      (self.vm.name, vendor))
        return vendor

    def get_vm_dmesg(self):
        """
        Get VM dmesg output.
        """
        cmd = "dmesg"
        return self.run_cmd(cmd)[1]

    def get_vm_parted(self):
        """
        Get vm parted info.
        """
        cmd = "parted -l"
        return self.run_cmd(cmd)[1]

    def get_vm_modules(self):
        """
        Get vm modules list.
        """
        cmd = "lsmod"
        return self.run_cmd(cmd)[1]

    def get_vm_pci_list(self):
        """
        Get vm pci list.
        """
        cmd_list = ['lspci', 'lshw', 'hwinfo']
        return self.run_cmd(cmd_list)[1]

    def get_vm_rc_local(self):
        """
        Get vm /etc/rc.local output.
        """
        cmd = "cat /etc/rc.local"
        return self.run_cmd(cmd)[1]

    def has_vmware_tools(self):
        """
        Check vmware tools.
        """
        cmd = "rpm -q VMwareTools"
        status, output = self.run_cmd(cmd)
        if status != 0:
            cmd = "ls /usr/bin/vmware-uninstall-tools.pl"
            status, output = self.run_cmd(cmd)
        return status == 0

    def get_vm_tty(self):
        """
        Get vm tty config.
        """
        confs = ('/etc/securetty', '/etc/inittab', '/boot/grub/grub.conf',
                 '/etc/default/grub')
        all_output = ''
        for conf in confs:
            cmd = "cat " + conf
            output = self.run_cmd(cmd)[1]
            all_output += output
        return all_output

    def wait_for_x_start(self, timeout=30):
        """
        Wait for S server start
        """
        cmd = 'xset -q'
        if self.run_cmd(cmd)[0] == 127:
            return
        utils_misc.wait_for(
            lambda: not bool(
                self.run_cmd(
                    cmd,
                    debug=False)[0]),
            timeout)

    def vm_xorg_search(self, substr):
        """
        Search expected string in Xorg config/log on VM.

        :param substr: The expected string.
        :return: True if search result meets expectation, otherwise False
        """
        self.wait_for_x_start()
        xorg_file_list = ["/etc/X11/xorg.conf",
                          "/var/log/Xorg.0.log"]
        # Ubuntu or rhel8 save xorg file in normal users home directory
        # A shell script gets the xorg file:
        #
        # uid_min=$(grep -E '^UID_MIN' /etc/login.defs | awk -F' ' '{ print $2}');
        # getent_cmd=('getent passwd' {${uid_min}..$(expr ${uid_min} + 100)});
        # for i in $(eval ${getent_cmd[@]} | awk -F: '{ print $1}');
        # do
        #    found=false;
        #    for j in $(seq 0 3);
        #    do
        #        if [ -f /home/${i}/.local/share/xorg/Xorg.${j}.log ]; then
        #            echo \"/home/${i}/.local/share/xorg/Xorg.${j}.log\";
        #            found=true;
        #            break;
        #        fi;
        #    done;
        #    if ${found}; then break; fi;
        # done
        get_uid_min = r"grep -E '^UID_MIN' /etc/login.defs | awk -F' ' '{ print $2}'"
        uid_min = self.run_cmd(get_uid_min, debug=False)[
            1].strip().splitlines()[-1]
        if uid_min.isdigit():
            # 100 times is enough
            uid_max = str(int(uid_min) + 100)
            extract_normal_users = r"getent passwd {%s..%s} | awk -F: '{ print $1}'" % (
                uid_min, uid_max)
            xorg_log_path = r"/home/${i}/.local/share/xorg/Xorg.${j}.log"
            xorg_log_chk = "if [ -f {0} ]; then echo {0}; found=true; break; fi;".format(
                xorg_log_path)
            break_if_found = r"if ${found}; then break; fi;"
            xorg_logs_loop = "for j in $(seq 0 3); do %s done; %s" % (
                xorg_log_chk, break_if_found)
            get_xorg_logs = "for i in $(%s);do found=false; %s done" % (
                extract_normal_users, xorg_logs_loop)
            logging.debug("Get xorg logs shell script:\n%s", get_xorg_logs)

            # The first element is a malformed get_xorg_logs string, it
            # should be removed.
            xorg_files = self.run_cmd(get_xorg_logs, debug=False)[
                1].strip().splitlines()
            if len(xorg_files) > 0:
                xorg_file_list.extend(xorg_files[1:])
        else:
            logging.debug("Get UID_MIN failed: %s", uid_min)

        logging.debug("xorg files: %s", xorg_file_list)
        for file_i in xorg_file_list:
            cmd = 'grep -i "%s" "%s"' % (substr, file_i)
            if self.run_cmd(cmd)[0] == 0:
                return True
        return False

    def vm_journal_search(self, substr, options=None, flags=re.IGNORECASE):
        """
        Search journalctl log on vm

        :param substr: The expected string.
        :param options: the journalctl options
        :param flags: A RegexFlag, Please refer RE moudule of python
        :return: True if search result meets expectation, otherwise False
        """
        cmd = "journalctl --no-pager"
        if options:
            cmd += " %s" % options

        if self.vm_general_search(
                cmd,
                substr,
                flags,
                ignore_status=True,
                debug=False):
            return True
        return False

    def vm_general_search(
            self,
            cmd,
            substr,
            flags=re.IGNORECASE,
            ignore_status=False,
            debug=True):
        """
        Search a string by running a command on vm

        :param cmd: A command to be excuted
        :param substr: The expected string in output of cmd.
        :param flags: A RegexFlag, Please refer RE moudule of python
        :param ignore_status: If True, will not check command return status.
        :param debug: If True, will print cmd output.
        :return: True if search result meets expectation, otherwise False
        """
        status, output = self.run_cmd(cmd, debug)

        if (ignore_status or status == 0) and re.search(substr, output, flags):
            return True
        return False

    def is_net_virtio(self):
        """
        Check whether vm's interface is virtio
        """
        cmd = "ls -l /sys/class/net/eth%s/device" % self.nic_index
        output = self.run_cmd(cmd)[1]
        try:
            if re.search("virtio", output.split('/')[-1]):
                return True
        except IndexError:
            logging.error("Fail to find virtio driver")
        return False

    def is_disk_virtio(self):
        """
        Check whether disk is virtio after convertion.

        Note: If kernel supports virtio_blk, v2v will always convert disks
        to virtio_blk in copying mode. That means all disk have /dev/vdx
        path name.
        """
        cmd = "fdisk -l"
        virtio_disks = r'/dev/vd[a-z]+[0-9]*'
        non_virtio_disks = r'/dev/[hs]d[a-z]+[0-9]*'
        output = self.run_cmd(cmd)[1]
        if re.search(non_virtio_disks, output):
            return False
        if re.search(virtio_disks, output):
            return True
        return False

    def is_uefi_guest(self):
        """
        Check whether guest is uefi guest
        """
        cmd = "ls /sys/firmware/efi"
        status, output = self.run_cmd(cmd)
        if status != 0:
            return False
        return True

    def get_grub_device(self, dev_map="/boot/grub*/device.map"):
        """
        Check whether vd[a-z] device is in device map.
        """
        cmd = "cat %s" % dev_map
        output = self.run_cmd(cmd)[1]
        if re.search("vd[a-z]", output):
            return True
        return False


class WindowsVMCheck(VMCheck):

    """
    This class handles all basic Windows VM check operations.
    """

    def send_win32_key(self, keycode):
        """
        Send key to Windows VM
        """
        options = "--codeset win32 %s" % keycode
        virsh.sendkey(self.name, options, session_id=self.virsh_session_id)
        time.sleep(1)

    def move_mouse(self, coordinate):
        """
        Move VM mouse.
        """
        virsh.move_mouse(self.name, coordinate,
                         session_id=self.virsh_session_id)

    def click_left_button(self):
        """
        Click left button of VM mouse.
        """
        virsh.click_button(self.name, session_id=self.virsh_session_id)

    def click_tab_enter(self):
        """
        Send TAB and ENTER to VM.
        """
        self.send_win32_key('VK_TAB')
        self.send_win32_key('VK_RETURN')

    def click_install_driver(self):
        """
        Move mouse and click button to install dirver for new
        device(Ethernet controller)
        """
        # Get window focus by click left button
        self.move_mouse((0, -80))
        self.click_left_button()
        self.move_mouse((0, 30))
        self.click_left_button()

    def get_screenshot(self):
        """
        Do virsh screenshot of the vm and fetch the image if the VM in
        remote host.
        """
        sshot_file = os.path.join(data_dir.get_tmp_dir(), "vm_screenshot.ppm")
        if self.target == "ovirt":
            # Note: This is a screenshot path on a remote host
            vm_sshot = os.path.join("/tmp", "vm_screenshot.ppm")
        else:
            vm_sshot = sshot_file
        virsh.screenshot(self.name, vm_sshot, session_id=self.virsh_session_id)
        if self.target == "ovirt":
            remote_ip = self.params.get("remote_ip")
            remote_user = self.params.get("remote_user")
            remote_pwd = self.params.get("remote_pwd")
            remote.scp_from_remote(remote_ip, '22', remote_user,
                                   remote_pwd, vm_sshot, sshot_file)
            r_runner = remote_old.RemoteRunner(
                host=remote_ip, username=remote_user, password=remote_pwd)
            r_runner.run("rm -f %s" % vm_sshot)
        return sshot_file

    def wait_for_match(self, images, similar_degree=0.98, timeout=300):
        """
        Compare VM screenshot with given images, if any image in the list
        matched, then return the image index, or return -1.
        """
        end_time = time.time() + timeout
        image_matched = False
        cropped_image = os.path.join(data_dir.get_tmp_dir(), "croped.ppm")
        while time.time() < end_time:
            vm_screenshot = self.get_screenshot()
            ppm_utils.image_crop_save(vm_screenshot, vm_screenshot)
            img_index = 0
            for image in images:
                logging.debug("Compare vm screenshot with image %s", image)
                ppm_utils.image_crop_save(image, cropped_image)
                h_degree = ppm_utils.image_histogram_compare(cropped_image,
                                                             vm_screenshot)
                if h_degree >= similar_degree:
                    logging.debug("Image %s matched", image)
                    image_matched = True
                    break
                img_index += 1
            if image_matched:
                break
            time.sleep(1)
        if os.path.exists(cropped_image):
            os.unlink(cropped_image)
        if os.path.exists(vm_screenshot):
            os.unlink(vm_screenshot)
        if image_matched:
            return img_index
        else:
            return -1

    def boot_windows(self, timeout=300):
        """
        Click buttons to activate windows and install ethernet controller driver
        to boot windows.
        """
        logging.info("Booting Windows in %s seconds", timeout)
        compare_screenshot_vms = ["win2003"]
        timeout_msg = "No matching screenshots found after %s seconds" % timeout
        timeout_msg += ", trying to log into the VM directly"
        match_image_list = []
        if self.os_version in compare_screenshot_vms:
            image_name_list = self.params.get("screenshots_for_match",
                                              '').split(',')
            for image_name in image_name_list:
                match_image = os.path.join(data_dir.get_data_dir(), image_name)
                if not os.path.exists(match_image):
                    logging.error(
                        "Screenshot '%s' does not exist", match_image)
                    return
                match_image_list.append(match_image)
            img_match_ret = self.wait_for_match(match_image_list,
                                                timeout=timeout)
            if img_match_ret < 0:
                logging.error(timeout_msg)
            else:
                if self.os_version == "win2003":
                    if img_match_ret == 0:
                        self.click_left_button()
                        # VM may have no response for a while
                        time.sleep(20)
                        self.click_left_button()
                        self.click_tab_enter()
                    elif img_match_ret == 1:
                        self.click_left_button()
                        time.sleep(20)
                        self.click_left_button()
                        self.click_tab_enter()
                        self.click_left_button()
                        self.send_win32_key('VK_RETURN')
                    else:
                        pass
                elif self.os_version in ["win7", "win2008r2"]:
                    if img_match_ret in [0, 1]:
                        self.click_left_button()
                        self.click_left_button()
                        self.send_win32_key('VK_TAB')
                        self.click_tab_enter()
                elif self.os_version == "win2008":
                    if img_match_ret in [0, 1]:
                        self.click_tab_enter()
                        self.click_install_driver()
                        self.move_mouse((0, -50))
                        self.click_left_button()
                        self.click_tab_enter()
                    else:
                        self.click_install_driver()
        else:
            # No need sendkey/click button for any os except Win2003
            logging.info("%s is booting up without program intervention",
                         self.os_version)

    def reboot_windows(self):
        """
        Reboot Windows immediately
        """
        cmd = "shutdown -t 0 -r -f"
        self.run_cmd(cmd)

    def get_viostor_info(self):
        """
        Get viostor info.
        """
        cmd = r"dir %s\Drivers\VirtIO\\viostor.sys" % self.windows_root
        return self.run_cmd(cmd)[1]

    def get_service_info(self, name=None):
        """
        Get service info.

        :param name: an optional service name
        """
        cmd = r"sc query"
        if name:
            cmd += ' ' + name
        return self.run_cmd(cmd)[1]

    def get_driver_info(self, signed=True):
        """
        Get windows signed driver info.
        """
        cmd = "DRIVERQUERY"
        if signed:
            cmd += " /SI"
        # Try 5 times to get driver info
        output, count = '', 5
        while count > 0:
            logging.debug('%d times remaining for getting driver info' % count)
            try:
                # Clean up output
                self.session.cmd('cls')
                output = self.session.cmd_output(cmd)
            except Exception as detail:
                logging.error(detail)
                count -= 1
            else:
                break
        if not output:
            logging.error('Fail to get driver info')
        logging.debug("Command output:\n%s", output)
        return output

    def get_windows_event_info(self):
        """
        Get windows event log info about WSH.
        """
        cmd = "wevtutil qe application | find \"WSH\""
        status, output = self.run_cmd(cmd)
        if status != 0:
            # For win2003 and winXP, use following cmd
            cmd = r"CSCRIPT %s\system32\eventquery.vbs " % self.windows_root
            cmd += "/l application /Fi \"Source eq WSH\""
            output = self.run_cmd(cmd)[1]
        return output

    def get_network_restart(self):
        """
        Get windows network restart.
        """
        cmd = "ipconfig /renew"
        return self.run_cmd(cmd)[1]

    def copy_windows_file(self):
        """
        Copy a widnows file
        """
        cmd = "COPY /y C:\\rss.reg C:\\rss.reg.bak"
        return self.run_cmd(cmd)[0]

    def delete_windows_file(self):
        """
        Delete a widnows file
        """
        cmd = "DEL C:\rss.reg.bak"
        return self.run_cmd(cmd)[0]

    def is_uefi_guest(self):
        """
        Check whether windows guest is uefi guest

        More info:
        https://www.tenforums.com/tutorials/85195-check-if-windows-10-using-uefi-legacy-bios.html
        """
        search_str = "Detected boot environment"
        target_file = r"c:\Windows\Panther\setupact.log"
        cmd = 'findstr /c:"%s" %s' % (search_str, target_file)
        status, output = self.run_cmd(cmd)
        if 'BIOS' in output:
            return False

        if 'EFI' in output:
            return True

        return False


def v2v_cmd(params, auto_clean=True, cmd_only=False, interaction=False):
    """
    Create finnal v2v command, execute or only return the command

    Sometimes you need to retouch the v2v command, then execute it later.
    So you need to preserve the resouces(nfs path, authorized keys, etc.

    When auto_clean is False, the resources created during runtime will
    not be cleaned up, Users should do that.

    When cmd_only is True, the v2v command will not be executed but be returned.
    Users can reedit the command as cases requried.

    :param params: A dictionary includes all of required parameters such as
                    'target', 'hypervisor' and 'hostname', etc.
                   This is a v2v specifc params and not the global params in
                   run function.
    :param auto_clean: boolean flag, whether to cleanup runtime resources.
    :param cmd_only: boolean flag, whether to only return the command line without running
    :param interaction: boolean flag, If need to interact with v2v
    :return: A cmd string or CmdResult object
    """
    def _v2v_pre_cmd():
        """
        Preprocess before running v2v cmd, such as starting VM for warm convertion,
        create virsh instance, etc.
        """
        # Cannot get mac address in 'ova', 'libvirtxml', etc.
        if input_mode not in [
            'disk',
            'libvirtxml',
            'local',
                'ova'] and not skip_virsh_pre_conn:
            params['_v2v_virsh'] = v2v_virsh = create_virsh_instance(
                hypervisor, uri, hostname, username, password)
            iface_info = get_all_ifaces_info(vm_name, v2v_virsh)
            # For v2v option '--mac', this is automatically generated.
            params['_iface_list'] = iface_info

            # Get disk count
            disk_count = vm_xml.VMXML.get_disk_count(
                vm_name, virsh_instance=v2v_virsh)
            params['_disk_count'] = disk_count

            if input_mode == 'vmx':
                disks_info = get_esx_disk_source_info(vm_name, v2v_virsh)
                if not disks_info:
                    raise exceptions.TestError(
                        "Found esx disk source error")
                # It's impossible that a VM is saved in two different
                # datastores
                params['datastore'] = list(disks_info)[0]
                params['_nfspath'] = list(disks_info[list(disks_info)[0]])[0]
        else:
            params['_iface_list'] = ''
            # Just set to 1 right now, but it could be improved if requried
            # in future
            params['_disk_count'] = 1
            # params['_nfspath'] only be used when composing nfs vmx file path,
            # in the case, vm_name is same as nfs directory name
            if input_mode == 'vmx':
                params['_nfspath'] = vm_name
        # Pass it to testcase to do subsequent checking
        global_params.update({'vm_disk_count': params['_disk_count']})

    def _v2v_post_cmd():
        """
        Postprocess after running v2v cmd
        """
        v2v_virsh = params.get('_v2v_virsh')
        close_virsh_instance(v2v_virsh)

    if V2V_EXEC is None:
        raise ValueError('Missing command: virt-v2v')

    global_params = params.get('params', {})
    if not global_params:
        # For the back compatibility reason, only report a warning message
        logging.warning(
            "The global params in run() need to be passed into v2v_cmd as an"
            "item of params, like {'params': params}. "
            "If not, some latest functions may not work as expected.")

    target = params.get('target')
    hypervisor = params.get('hypervisor')
    # vpx:// or esx://
    src_uri_type = params.get('src_uri_type')
    hostname = params.get('hostname')
    vpx_dc = params.get('vpx_dc')
    esxi_host = params.get('esxi_host', params.get('esx_ip'))
    opts_extra = params.get('v2v_opts')
    # Set v2v_cmd_timeout to 5 hours, the value can give v2v enough time to execute,
    # and avoid v2v process be killed by mistake.
    # the value is bigger than the timeout value in CI, so when some timeout
    # really happens, CI will still interrupt the v2v process.
    v2v_cmd_timeout = params.get('v2v_cmd_timeout', 18000)
    # username and password of remote hypervisor server
    username = params.get('username', 'root')
    password = params.get('password')
    vm_name = params.get('main_vm')
    input_mode = params.get('input_mode')
    cmd_has_ip = params.get('cmd_has_ip', True)
    # A switch controls a virsh pre-connection to source hypervisor,
    # but some testing envrionments(like, gating in OSP) don't have
    # source hypervisor, the pre-connection must be skipped.
    skip_virsh_pre_conn = 'yes' == params.get('skip_virsh_pre_conn')
    # virsh instance of remote hypervisor
    params.update({'_v2v_virsh': None})

    uri_obj = Uri(hypervisor)

    # Return actual 'uri' according to 'hostname' and 'hypervisor'
    if src_uri_type == 'esx':
        vpx_dc = None
    uri = uri_obj.get_uri(hostname, vpx_dc, esxi_host)

    try:
        # Pre-process for v2v
        _v2v_pre_cmd()

        target_obj = Target(target, uri)
        # Return virt-v2v command line options based on 'target' and
        # 'hypervisor'
        options = target_obj.get_cmd_options(params)

        if opts_extra:
            options = options + ' ' + opts_extra
        # Add -oo rhv-disk-uuid if '--no-copy' exists
        if '-o rhv-upload' in options and '--no-copy' in options and '--oo rhv-disk-uuid' not in options:
            for i in range(int(params.get('_disk_count', 0))):
                options += ' -oo rhv-disk-uuid=%s' % str(uuid.uuid4())

        # Construct a final virt-v2v command and remove redundant blanks
        cmd = ' '.join(('%s %s' % (V2V_EXEC, options)).split())
        # Old v2v version doesn't support '-ip' option
        if not v2v_supported_option("-ip <filename>"):
            cmd = cmd.replace('-ip', '--password-file', 1)

        # update -ip option
        if not cmd_has_ip:
            ip_ptn = [r'-ip \S+\s*', r'--password-file \S+\s*']
            cmd = cmd_remove_option(cmd, ip_ptn)
        # Save v2v command to params, then it can be passed to
        # import_vm_to_ovirt
        global_params.update({'v2v_command': cmd})

        if not cmd_only:
            if not interaction:
                cmd_result = process.run(
                    cmd,
                    timeout=v2v_cmd_timeout,
                    verbose=True,
                    ignore_status=True)
            else:
                cmd_result = interactive_run(
                    params,
                    timeout=v2v_cmd_timeout,
                    command=cmd,
                    output_func=lambda x: print(x, end=''))
    finally:
        # Save it into global params and release it by users
        v2v_dirty_resources = global_params.get('v2v_dirty_resources', [])
        global_params.update({'v2v_dirty_resources': v2v_dirty_resources})
        if 'target_obj' in locals():
            if auto_clean:
                target_obj.cleanup()
            else:
                v2v_dirty_resources.append(target_obj)
        # Post-process for v2v
        _v2v_post_cmd()

    if cmd_only:
        return cmd
    cmd_result.stdout = cmd_result.stdout_text
    cmd_result.stderr = cmd_result.stderr_text
    return cmd_result


def cmd_run(cmd, obj_be_cleaned=None, auto_clean=True, timeout=18000):
    """
    Run v2v command and cleanup resources automatically

    When you preserved the resources in v2v_cmd, the users need to release
    them. This function can help you do it automatically.

    :param obj_be_cleaned: the resources to be cleaned up
    :param auto_clean: If true, cleanup obj_be_cleaned automatically
    :param timeout: the timeout value for the command to run
    """
    try:
        cmd_result = process.run(
            cmd,
            timeout=timeout,
            verbose=True,
            ignore_status=True)
    finally:
        if auto_clean and obj_be_cleaned:
            logging.debug('Running cleanup for %s', obj_be_cleaned)
            if isinstance(obj_be_cleaned, list):
                for obj in obj_be_cleaned:
                    obj.cleanup()
            else:
                obj_be_cleaned.cleanup()

    return cmd_result


def import_vm_to_ovirt(params, address_cache, timeout=600):
    """
    Import VM from export domain to oVirt Data Center
    """
    v2v_cmd = params.get('v2v_command')
    vm_name = params.get('main_vm')
    os_type = params.get('os_type')
    export_name = params.get('export_name')
    storage_name = params.get('storage_name')
    cluster_name = params.get('cluster_name')
    output_method = params.get('output_method')
    # Check oVirt status
    dc = ovirt.DataCenterManager(params)
    logging.info("Current data centers list: %s", dc.list())
    cm = ovirt.ClusterManager(params)
    logging.info("Current cluster list: %s", cm.list())
    hm = ovirt.HostManager(params)
    logging.info("Current host list: %s", hm.list())
    sdm = ovirt.StorageDomainManager(params)
    logging.info("Current storage domain list: %s", sdm.list())
    vm = ovirt.VMManager(vm_name, params, address_cache=address_cache)
    logging.info("Current VM list: %s", vm.list())
    if vm_name in vm.list() and output_method != 'rhv_upload':
        logging.error("%s already exist", vm_name)
        return False
    wait_for_up = True
    if os_type == 'windows':
        wait_for_up = False

    # If output_method is None or "" or is not 'rhv_upload', treat it as
    # old way.
    if output_method != 'rhv_upload':
        try:
            # Import VM
            vm.import_from_export_domain(export_name,
                                         storage_name,
                                         cluster_name,
                                         timeout=timeout)
            logging.info("The latest VM list: %s", vm.list())
        except Exception as e:
            # Try to delete the vm from export domain
            vm.delete_from_export_domain(export_name)
            logging.error("Import %s failed: %s", vm.name, e)
            return False
    try:
        if not is_option_in_v2v_cmd(v2v_cmd, '--no-copy'):
            # Start VM
            vm.start(wait_for_up=wait_for_up)
        else:
            logging.debug(
                'Skip starting VM: --no-copy is in cmdline:\n%s',
                v2v_cmd)
    except Exception as e:
        logging.error("Start %s failed: %s", vm.name, e)
        vm.delete()
        if output_method != 'rhv_upload':
            vm.delete_from_export_domain(export_name)
        return False
    return True


def check_log(params, log):
    """
    Check if error/warning meets expectation in v2v log

    :param params: A dictionary includes all of required parameters such as
                    'expect_msg', 'msg_content', etc.
    :param log: The log string to be checked
    :return: True if search result meets expectation, otherwise False
    """

    def _check_log(pattern_list, expect=True):
        for pattern in pattern_list:
            line = r'\s*'.join(pattern.split())
            expected = 'expected' if expect else 'not expected'
            logging.info('Searching for %s log: %s' % (expected, pattern))
            compiled_pattern = re.compile(line, flags=re.S)
            search = re.search(compiled_pattern, log)
            if search:
                logging.info('Found log: %s', search.group(0))
                if not expect:
                    return False
            else:
                logging.info('Not find log: %s', pattern)
                if expect:
                    return False
        return True

    expect_msg = params.get('expect_msg')
    ret = ''
    if not expect_msg:
        logging.info('No need to check v2v log')
    else:
        expect = expect_msg == 'yes'
        if params.get('msg_content'):
            msg_list = params['msg_content'].split('%')
            if _check_log(msg_list, expect=expect):
                logging.info('Finish checking v2v log')
            else:
                ret = 'Check v2v log failed'
        else:
            ret = 'Missing error message to compare'
    return ret


def check_exit_status(result, expect_error=False, error_flag='strict'):
    """
    Check the exit status of virt-v2v/libguestfs commands

    :param result: Virsh command result object
    :param expect_error: Boolean value, expect command success or fail
    :param error_flag: same as errors argument in str.decode
    """
    if not expect_error:
        if result.exit_status != 0:
            raise exceptions.TestFail(
                to_text(result.stderr, errors=error_flag))
    elif expect_error and result.exit_status == 0:
        raise exceptions.TestFail("Run '%s' expect fail, but run "
                                  "successfully." % result.command)


def cleanup_constant_files(params):
    """
    Cleanup some constant files which generated for v2v commands.
    For example, rhv_upload_passwd_file, local_ca_file_path,
    vpx_passwd_file, etc.

    :param params: A dict containing all cfg params
    """
    # Please Add new constant files into below list.
    tmpfiles = [params.get("rhv_upload_passwd_file"),
                params.get("local_ca_file_path"),
                params.get("vpx_passwd_file")]

    # Python3 only returns a map object which is different from python2.
    list(map(os.remove, [x for x in tmpfiles if x and os.path.isfile(x)]))


def get_vddk_thumbprint(host, password, uri_type, prompt=r"[\#\$\[\]]"):
    """
    Get vddk thumbprint from VMware vCenter

    :param host: hostname or IP address
    :param password: Password
    :param uri_type: conversion source uri type
    :param prompt: Shell prompt (regular expression)
    """

    if uri_type == 'esx':
        cmd = 'openssl x509 -in /etc/vmware/ssl/rui.crt -fingerprint -sha1 -noout'
    else:
        cmd = 'openssl x509 -in /etc/vmware-vpx/ssl/rui.crt -fingerprint -sha1 -noout'

    r_runner = remote_old.RemoteRunner(
        host=host,
        password=password,
        prompt=prompt,
        preferred_authenticaton='password,keyboard-interactive')
    cmdresult = r_runner.run(cmd)
    logging.debug("vddk thumbprint:\n%s", cmdresult.stdout)
    vddk_thumbprint = cmdresult.stdout.strip().split('=')[1]

    return vddk_thumbprint


def v2v_setup_ssh_key(
        hostname,
        username,
        password,
        port=22,
        server_type=None,
        auto_close=True,
        preferred_authenticaton=None,
        user_known_hosts_file=None):
    """
    Setup up remote login in another server by using public key

    :param hostname: hostname or IP address
    :param username: username
    :param password: password
    :param port: ssh port number
    :param server_type: the type of remote server, the values could be 'esx' or 'None'.
    :param auto_close: If it's True, the session will closed automatically,
                       else Uses should call v2v_setup_ssh_key_cleanup to close the session
    :param preferred_authenticaton: The preferred authentication of SSH connection
    :param user_known_hosts_file: one or more files to use for the user host key database

    :return: A tuple (public_key, session) will always be returned
    """
    session = None
    logging.debug('Performing SSH key setup on %s:%d as %s.' %
                  (hostname, port, username))
    try:
        # Both Xen and ESX can work with following settings.
        if not preferred_authenticaton:
            preferred_authenticaton = 'password,keyboard-interactive'
        if not user_known_hosts_file:
            user_known_hosts_file = os.path.expanduser('~/.ssh/known_hosts')

        # If remote host identification has changed, v2v will fail.
        # We always delete the identification first.
        if os.path.isfile(user_known_hosts_file):
            cmd = r"sed -i '/%s/d' %s" % (hostname, user_known_hosts_file)
            process.run(cmd, verbose=True, ignore_status=True)

        session = remote_old.ssh_login_to_migrate(
            client='ssh',
            host=hostname,
            username=username,
            port=port,
            password=password,
            prompt=r"[\#\$\[\]%]",
            verbose=True,
            preferred_authenticaton=preferred_authenticaton,
            user_known_hosts_file=user_known_hosts_file)

        # Add rstrip to avoid blank lines in authorized_keys on remote server
        public_key = ssh_key.get_public_key().rstrip()

        if server_type == 'esx':
            session.cmd(
                "echo '%s' >> /etc/ssh/keys-root/authorized_keys; " %
                public_key)
        else:
            session.cmd('mkdir -p ~/.ssh')
            session.cmd('chmod 700 ~/.ssh')
            session.cmd("echo '%s' >> ~/.ssh/authorized_keys; " % public_key)
            session.cmd('chmod 600 ~/.ssh/authorized_keys')

        logging.debug('SSH key setup complete, session is %s', session)

        return public_key, session
    except Exception as err:
        # auto close session when exception occurs
        auto_close = True
        raise exceptions.TestFail("SSH key setup failed: '%s'" % err)
    finally:
        if auto_close and session:
            logging.debug('cleaning session: %s', session)
            session.close()


def v2v_setup_ssh_key_cleanup(session=None, key=None, server_type=None):
    """
    Close the session and delete the key from authorized_keys.

    If auto_close is 'False' in v2v_setup_ssh_key, Users must call this
    function to cleanup session and keys on remote server explicitly.
    But if the caller of v2v_setup_ssh_key is an instance of class Target,
    you can save the resources in self.authorized_keys to cleanup
    automatically.

    :param session: An aexpect session to remote server
    :param key: The public_key which get by ssh_key.get_public_key().
    :param server_type: the type of remote server, the values could be 'esx' or 'None'.
    """
    try:
        if not session or not key:
            return

        # Only use a part of pub_keys as a pattern in sed
        key = key.rstrip().split()[1].split('/')[0]
        authorized_keys = get_authorized_keys_file(server_type)
        cmd = r"sed -i '/%s/d' %s" % (key, authorized_keys)
        session.cmd(cmd)
    finally:
        if session:
            logging.debug('cleaning session: %s', session)
            session.close()


def get_authorized_keys_file(server_type=None):
    """
    Get authorized_keys file path for a remote server

    :param server_type: the type of remote server, the values could be 'esx' or 'None'.

    :return: The path of authorized_keys file on remote server
    """
    if server_type == 'esx':
        authorized_keys = '/etc/ssh/keys-root/authorized_keys'
    else:
        authorized_keys = os.path.expanduser('~/.ssh/authorized_keys')
    return authorized_keys


def v2v_mount(src, dst='v2v_mount_point', fstype='nfs'):
    """
    Mount nfs src to dst

    :param src: NFS source
    :param dst: NFS mount point
    """
    mount_point = os.path.join(
        data_dir.get_tmp_dir(), dst)
    if not os.path.exists(mount_point):
        os.makedirs(mount_point)

    if not utils_misc.mount(
        src,
        mount_point,
        fstype,
            verbose=True):
        raise exceptions.TestError(
            'Mount %s for %s failed' %
            (src, mount_point))

    return mount_point


def create_virsh_instance(
        hypervisor,
        uri,
        remote_ip,
        remote_user,
        remote_pwd,
        debug=True):
    """
    Create a virsh instance for all hypervisors(VMWARE, XEN, KVM)

    :param hypervisor: a hypervisor type
    :param uri: uri of libvirt instance to connect to
    :param remote_ip: Hostname/IP of remote system to ssh into (if any)
    :param remote_user: Username to ssh in as (if any)
    :param remote_pwd: Password to use, or None for host/pubkey
    :param debug: Whether to enable debug
    """
    logging.debug(
        "virsh connection info: hypervisor=%s uri=%s ip=%s",
        hypervisor,
        uri,
        remote_ip)
    if hypervisor == 'kvm':
        v2v_virsh = virsh
    else:
        virsh_dargs = {'uri': uri,
                       'remote_ip': remote_ip,
                       'remote_user': remote_user,
                       'remote_pwd': remote_pwd,
                       'auto_close': True,
                       'debug': debug}
        v2v_virsh = wait_for(virsh.VirshPersistent, **virsh_dargs)
    logging.debug('A new virsh persistent session %s was created', v2v_virsh)
    return v2v_virsh


def close_virsh_instance(virsh_instance=None):
    """
    Close a virsh instance

    :param v2v_virsh_instance: a virsh instance
    """

    logging.debug('Closing session (%s) in VT', virsh_instance)
    if virsh_instance and hasattr(virsh_instance, 'close_session'):
        virsh_instance.close_session()


def get_all_ifaces_info(vm_name, virsh_instance):
    """
    Get all interfaces' information

    :param vm_name: vm's name
    :param v2v_virsh_instance: a virsh instance
    """
    # virsh can't find guest every time on old XEN server.
    vmxml = wait_for(
        vm_xml.VMXML.new_from_dumpxml,
        vm_name=vm_name,
        virsh_instance=virsh_instance)
    interfaces = vmxml.get_iface_all()
    if len(interfaces.keys()) == 0:
        raise exceptions.TestError("Not found mac address for vm %s" % vm_name)

    # vm_ifaces = [(mac, type), ...]
    vm_ifaces = []
    for mac, iface in interfaces.items():
        vm_ifaces.append((mac, iface.get('type')))

    logging.debug("Iface information for vm %s: %s", vm_name, vm_ifaces)
    return vm_ifaces


def get_esx_disk_source_info(vm_name, virsh_instance):
    """
    Get VM's datastore, real path of disks in NFS, etc

    The return values looks like:
    {'esx6.7': {'esx6.7-rhel7.5-multi-disks':
      ['esx6.7-rhel7.5-multi-disks.vmdk', 'esx6.7-rhel7.5-multi-disks_2.vmdk']}}

    It can be used to construct real path value in
    '-i vmx -it ssh' mode of v2v command.

    disk info example:
    <disk type='file' device='disk'>
      <source file='[esx6.7] esx6.7-rhel7.5-multi-disks/esx6.7-rhel7.5-multi-disks.vmdk'/>
      <target dev='sda' bus='scsi'/>
      <address type='drive' controller='0' bus='0' target='0' unit='0'/>
    </disk>

    The meanings of source file info.
    '[esx6.7]' is datastore of the VM.
    'esx6.7-rhel7.5-multi-disks' is the real path in NFS storage.
    'esx6.7-rhel7.5-multi-disks.vmdk' is disk name.

    Only works for VMs in ESX server

    :param vm_name: vm's name
    :param v2v_virsh_instance: a virsh instance
    """
    def _parse_file_info(path):
        res = re.search(r'\[(.*)\] (.*?)/(.*\.vmdk)', path)
        if not res:
            return []
        return [res.group(i) for i in range(1, 4)]

    disks_info = {}
    disks = vm_xml.VMXML.get_disk_source(
        vm_name, virsh_instance=virsh_instance)
    for disk in disks:
        attr_value = disk.find('source').get('file')
        file_info = _parse_file_info(attr_value)
        if not file_info:
            continue
        if file_info[0] not in disks_info:
            disks_info[file_info[0]] = {file_info[1]: []}
        if file_info[1] not in disks_info[file_info[0]]:
            disks_info[file_info[0]][file_info[1]] = []
        disks_info[file_info[0]][file_info[1]].append(file_info[2])

    logging.debug("source disk info vm %s: %s", vm_name, disks_info)
    return disks_info


def v2v_supported_option(opt_str):
    """
    Check if an cmd option is supported by v2v

    :param opt_str: option string for checking
    """
    cmd = 'virt-v2v --help'
    result = process.run(cmd, verbose=False, ignore_status=True)
    if re.search(r'%s' % opt_str, result.stdout_text):
        return True
    return False


def is_option_in_v2v_cmd(v2v_cmd, option):
    """
    Check if specific v2v option(s) is(are) in v2v command line

    :param v2v_cmd: A v2v command line string
    :param option: option or option list
    """
    if not option or not v2v_cmd:
        return False

    if isinstance(option, list):
        return all([opt_i in v2v_cmd for opt_i in option])
    elif isinstance(option, str):
        return option in v2v_cmd
    else:
        return False


def wait_for(func, timeout=300, interval=10, *args, **kwargs):
    """
    Run a function 'func' until success or timeout

    :param func: the function to be called
    :param timeout: timeout value(seconds)
    :param interval: interval values (seconds) for every call
    :param *args: arguments to be passed to func
    :param **dwargs: diction arguments to be passed to func
    """
    count = 0
    end_time = time.time() + float(timeout)
    while time.time() < end_time:
        try:
            return func(*args, **kwargs)
        except Exception:
            count += 1
            time.sleep(interval)

    logging.debug("Tried %s times", count)
    # Run once more, raise exception or success
    return func(*args, **kwargs)


def check_version(version, interval):
    """
    Check version against given interval string.

    :param version: The version to be compared
    :param interval: An interval is a string representation of a
     mathmetical like interval. See the defination in utils_version.py.
    :return: True if version satisfied interval, otherwise False.
    """
    verison_interval = VersionInterval(interval)
    return version in verison_interval


def compare_version(interval, version=None, cmd=None):
    """
    Compare version against given interval string.

    :param interval: An interval is a string representation of a
     mathmetical like interval. See the defination in utils_version.py.
    :param version: The version to be compared
    :param cmd: the command to get the version
    :return: True if version satisfied interval, otherwise False.
    """
    if not version:
        if not cmd:
            cmd = 'rpm -q virt-v2v'
        # exit status checking will be done in process.run
        version = process.run(cmd, shell=True).stdout_text.strip()

    return check_version(version, interval)


def multiple_versions_compare(interval):
    """
    Mutiple pkgs can be specified by ';', e.g.
    "[libguestfs-1.40,);[nbkdit-1.17.4,)"

    If interval is '', it means no version limitation and return True.
    If interval is not '', return True for compareing success and False
    for Failure.

    :param interval: An interval is a string representation of a
    """
    re_pkg_name = r'(.*?)-(?=\d+\.?)+'
    versions = interval.split(';')
    # ';' is used to split mutiple pkgs.
    for ver_i in versions:
        ver = ver_i.strip('[]()')
        if not ver:
            continue
        if not re.search(re_pkg_name, ver):
            return False

        pkg_name = re.search(re_pkg_name, ver).group(1)
        cmd = 'rpm -q %s' % pkg_name
        if not compare_version(ver_i, cmd=cmd):
            return False

    return True


def to_list(value):
    """
    Convert a string to list

    :param value: A string or list object
    """
    if not value:
        return []
    elif isinstance(value, list):
        return value
    elif isinstance(value, str):
        return [value]


def cmd_remove_option(cmd, opt_pattern):
    """
    Remove an option from cmd

    :param cmd: the cmd
    :param opt_pattern: a pattern stands for the option
    """
    for pattern in to_list(opt_pattern):
        for item in re.findall(pattern, cmd):
            cmd = cmd.replace(item, '').strip()
    return cmd


def params_get(params, name, default=None):
    """
    A convient function for v2v to get value of a variant from params.

    The main advantange is all variants don't need to be passed to
    utils_v2v any more, this function will get it from the standard
    'params' if 'params' was passed to utils_v2v.

    This requires all variants have same name in the whole test life.

    :param params: A dictionary includes all of required parameters.
    :param name: A variant name
    :param default: The default value of a variant
    """
    return params.get(name) or (params.get('params').get(
        name, default) if 'params' in params else params.get(name, default))


def interactive_run(params, timeout=300, *args, **kwargs):
    """
    Run v2v command interactively.

    :param params: A dictionary includes all of required parameters.
    :param timeout: The max command running time.
    """
    username = params_get(params, 'username', 'root')
    password = params_get(params, 'password')
    luks_password = params_get(params, 'luks_password')
    choices = params_get(params, 'custom_inputs')
    cmd_result = process.CmdResult()

    # For last line matching
    LAST_LINE_PROMPTS = [r"[Ee]nter.*username",
                         r"[Ee]nter.*authentication name",
                         r"[Ee]nter root.*? password",
                         r"[Ee]nter host password",
                         r"[Ee]nter.*password",
                         r"password:",
                         r"Enter a number between 1 and 2",
                         r"Enter key or passphrase",
                         r"Finishing off",
                         r"Converting .*? to run on"]

    # Interaction Done
    FREE_RUNNING_PROMPTS = [r"Converting .*? to run on"]

    def handle_prompts(session, timeout=300, interval=1.0):
        """
        Interact with v2v when specific patterns are matched.

        :param session: An Expect or ShellSession instance to operate on
        :param timeout: The max command running time.
        :param interval: Seconds to wait until start reading output again
        """

        free_running = False
        last_line = ''
        # v2v running timeout, it is set when interaction finished.
        # This value gives v2v enough time to execute.
        # If '-v -x' not enabled, it should be greater than 7200s.
        running_timeout = timeout
        # interactive timeout
        # when debug enabled in v2v, lots of logs keep outputing to stdout,
        # the timeout value can be smaller. If debug is off, the timeout
        # should be big enough to avoid unexpected timeout.
        timeout = 120 if '-v -x' in session.command else 300
        while True:
            time.sleep(interval)
            if not session.is_alive() or session.is_defunct():
                break

            try:
                match, _ = session.read_until_last_line_matches(
                    LAST_LINE_PROMPTS, timeout=timeout, internal_timeout=0.5)
                if match in [0, 1]:  # "username:"
                    logging.debug(
                        "Got username prompt; sending '%s'", username)
                    session.sendline(username)
                elif match in [2, 3, 4, 5]:
                    logging.debug(
                        "Got password prompt, sending '%s'",
                        asterisk_passwd(password))
                    session.sendline(password)
                elif match == 6:  # Wait for custom input
                    logging.debug(
                        "Got console '%s', send input list %s", match, choices)
                    session.sendline(choices)
                elif match == 7:  # LUKS password
                    logging.debug(
                        "Got password prompt, sending '%s'",
                        asterisk_passwd(luks_password))
                    session.sendline(luks_password)
                elif match == 8:  # Done
                    break
                elif match == 9:  # Interaction Finish
                    timeout = running_timeout
            except aexpect.ExpectTimeoutError:
                # when free_running is true and timeout happens, it means
                # the command doesn't have any response, may be dead or
                # performance is quite poor.
                logging.debug("timeout happens")
                if free_running:
                    raise

                # If timeout happens two times and the last line are same,
                # it means v2v may be dead or performance is quite poor.
                cont = session.get_output()
                new_last_line = ''
                nonempty_lines = [
                    l for l in cont.splitlines()[-10:] if l.strip()]
                if nonempty_lines:
                    new_last_line = nonempty_lines[-1]
                if last_line and last_line == new_last_line:
                    logging.debug(
                        'v2v command may be dead or have bad performance')
                    raise
                last_line = new_last_line

                # Set a big timeout value when interaction finishes
                for pattern in FREE_RUNNING_PROMPTS:
                    if re.search(pattern, cont):
                        logging.debug(
                            "interaction finished and begin running freely")
                        free_running = True
                        timeout = running_timeout
                        break

    try:
        subproc = aexpect.Expect(*args, **kwargs)
        logging.debug('Running command: %s', subproc.command)
        handle_prompts(subproc, timeout)
    except aexpect.ExpectProcessTerminatedError:
        # v2v cmd is dead or finished
        pass
    except Exception:
        # aexpect.ExpectTimeoutError or other exceptions
        # send 'ctrl+c' to v2v process to interrupt running quickly
        subproc.sendcontrol('c')
        raise
    finally:
        logging.debug(
            "Command '%s' finished with status %s",
            subproc.command,
            subproc.get_status())
        # Set command result
        cmd_result.command = subproc.command
        cmd_result.exit_status = subproc.get_status()
        cmd_result.stdout = subproc.get_output()
        # Close the expect session
        subproc.close()

    return cmd_result



import threading


from aexpect import remote


class ConsoleError(Exception):
    """A base class for console errors."""
    pass


class NoConsoleError(ConsoleError):
    """Raise when there is no console available for a session."""

    pass


class ConsoleBusyError(ConsoleError):
    """Raise when session tries to communicate with a console in use."""

    pass


def lock(function):
    """
    Get the ConsoleManager lock, run the function, then release the lock.

    :param function: Function to wrap.
    """
    def wrapper(*args, **kwargs):
        console_manager = args[0]
        if not console_manager._lock.acquire(False):
            raise ConsoleBusyError("Console is in use.")
        try:
            return function(*args, **kwargs)
        finally:
            console_manager._lock.release()
    return wrapper


class ConsoleManager(object):
    """A class for console session communication pipeline."""

    def __init__(self):
        self._console = None
        self.status_test_command = None
        self._lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_lock"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._lock = threading.Lock()

    @lock
    def __login(self, linesep, status_test_command,
                prompt, username, password, timeout):
        self._console.set_linesep(linesep)
        self._console.set_status_test_command(status_test_command)
        remote.handle_prompts(self._console, username, password,
                              prompt, timeout)

    def create_session(self, linesep, status_test_command,
                       prompt, username, password, timeout):
        """
        Return a console session with itself as the manager.

        :param linesep: line separator for console
        :param status_test_command: status test command for console
        :param prompt: console prompt pattern
        :param username: login username
        :param password: login password
        :param timeout: time (seconds) before giving up logging into the guest
        """
        if self._console is None:
            raise NoConsoleError("No console available.")
        self.__login(linesep, status_test_command, prompt,
                     username, password, timeout)
        return ConsoleSession(self)

    def set_console(self, console):
        self._console = console
        if self._console is not None:
            self.status_test_command = self._console.status_test_command

    @lock
    def proxy_call(self, func, *args, **kwargs):
        """Proxy function call to call functions provided by a Console."""

        _func = getattr(self._console, func)
        return _func(*args, **kwargs)


class ConsoleSession(object):
    """
    A wrapper for communicating with console.

    For more detailed function call information,
    you may need to refer to aexpect.client.ShellSession().
    """

    def __init__(self, manager):
        self.__manager = manager
        self.status_test_command = manager.status_test_command
        self.__closed = False

    def __repr__(self):
        return "<console session %s>" % id(self)

    def __verify_session_status(self):
        """Check the session status,if it is closed, raise error."""
        if self.__closed:
            raise RuntimeError("%s is closed." % self)

    def is_responsive(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.is_responsive.__name__,
                                         *args, **kwargs)

    def cmd_output(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.cmd_output.__name__,
                                         *args, **kwargs)

    def cmd_output_safe(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.cmd_output_safe.__name__,
                                         *args, **kwargs)

    def cmd_status_output(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.cmd_status_output.__name__,
                                         *args, **kwargs)

    def cmd_status(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.cmd_status.__name__,
                                         *args, **kwargs)

    def cmd(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.cmd.__name__,
                                         *args, **kwargs)

    def close(self):
        self.__verify_session_status()
        self.__closed = True

# FIXME: the following methods are temporarily introduced to workaround
#        console-session issues caused by the incorrect usages

    def send(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.send.__name__,
                                         *args, **kwargs)

    def sendline(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.sendline.__name__,
                                         *args, **kwargs)

    def sendcontrol(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.sendcontrol.__name__,
                                         *args, **kwargs)

    def send_ctrl(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.send_ctrl.__name__,
                                         *args, **kwargs)

    def set_linesep(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.set_linesep.__name__,
                                         *args, **kwargs)

    def read_nonblocking(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.read_nonblocking.__name__,
                                         *args, **kwargs)

    def read_until_output_matches(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(
            self.read_until_output_matches.__name__, *args, **kwargs)

    def read_until_last_line_matches(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(
            self.read_until_last_line_matches.__name__, *args, **kwargs)

    def read_until_any_line_matches(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(
            self.read_until_any_line_matches.__name__, *args, **kwargs)

    def read_up_to_prompt(self, *args, **kwargs):
        self.__verify_session_status()
        return self.__manager.proxy_call(self.read_up_to_prompt.__name__,
                                         *args, **kwargs)



import errno
import logging
import os
import re
import random
import string
import sys

from avocado.core import exit_codes
from avocado.utils.process import pid_exists
from avocado.utils.stacktrace import log_exc_info

from avocado.core.plugin_interfaces import JobPreTests as Pre
from avocado.core.plugin_interfaces import JobPostTests as Post

from ..test import VirtTest

from virttest.compat import get_settings_value

from six.moves import xrange


class LockCreationError(Exception):

    """
    Represents any error situation when attempting to create a lock file
    """
    pass


class OtherProcessHoldsLockError(Exception):

    """
    Represents a condition where other process has the lock
    """


class VTJobLock(Pre, Post):

    name = 'vt-joblock'
    description = 'Avocado-VT Job Lock/Unlock'

    def __init__(self, **kwargs):
        self.log = logging.getLogger("avocado.app")
        lock_dir = get_settings_value("plugins.vtjoblock", "dir",
                                      key_type=str, default='/tmp')
        self.lock_dir = os.path.expanduser(lock_dir)
        self.lock_file = None

    def _create_self_lock_file(self, job):
        """
        Creates the lock file for this job process

        :param job: the currently running job
        :type job: :class:`avocado.core.job.Job`
        :raises: :class:`LockCreationError`
        :returns: the full path for the lock file created
        :rtype: str
        """
        pattern = 'avocado-vt-joblock-%(jobid)s-%(uid)s-%(random)s.pid'
        # the job unique id is already random, but, let's add yet another one
        rand = ''.join([random.choice(string.ascii_lowercase + string.digits)
                        for i in xrange(8)])
        path = pattern % {'jobid': job.unique_id,
                          'uid': os.getuid(),
                          'random': rand}
        path = os.path.join(self.lock_dir, path)

        try:
            with open(path, 'w') as lockfile:
                lockfile.write("%u" % os.getpid())
            return path
        except Exception as e:
            raise LockCreationError(e)

    def _get_lock_files(self):
        """
        Get the list of lock file names under the current lock dir

        :returns: a list with the full path of files that match the
                  lockfile pattern
        :rtype: list
        """
        try:
            files = os.listdir(self.lock_dir)
            pattern = re.compile(r'avocado-vt-joblock-[0-9a-f]{40}-[0-9]+'
                                 '-[0-9a-z]{8}\.pid')
            return [os.path.join(self.lock_dir, _) for _ in files
                    if pattern.match(_)]
        except OSError as e:
            if e.errno == errno.ENOENT:
                return []

    def _lock(self, job):
        self.lock_file = self._create_self_lock_file(job)
        lock_files = self._get_lock_files()
        lock_files.remove(self.lock_file)
        for path in lock_files:
            try:
                lock_pid = int(open(path, 'r').read())
            except Exception:
                msg = 'Cannot read PID from "%s".' % path
                raise LockCreationError(msg)
            else:
                if pid_exists(lock_pid):
                    msg = 'File "%s" acquired by PID %u. ' % (path, lock_pid)
                    raise OtherProcessHoldsLockError(msg)
                else:
                    try:
                        os.unlink(path)
                    except OSError:
                        self.log.warn("Unable to remove stale lock: %s", path)

    @staticmethod
    def _get_klass_or_none(test_factory):
        try:
            return test_factory[0]
        except TypeError:
            return None

    def pre_tests(self, job):
        try:
            if job.test_suite is not None:
                if hasattr(job.test_suite, 'tests'):
                    tests = job.test_suite.tests
                else:
                    tests = job.test_suite
                if any(self._get_klass_or_none(test_factory) is VirtTest
                       for test_factory in tests):
                    self._lock(job)
        except Exception as detail:
            msg = "Failure trying to set Avocado-VT job lock: %s" % detail
            self.log.error(msg)
            log_exc_info(sys.exc_info(), self.log.name)
            sys.exit(exit_codes.AVOCADO_JOB_FAIL | job.exitcode)

    def post_tests(self, job):
        if self.lock_file is not None:
            os.unlink(self.lock_file)

    # compatibility with older versions, including 36.0 LTS
    pre = pre_tests
    post = post_tests



"""
Specializations of base.AccessorBase for particular XML manipulation types
"""

import logging

from virttest import xml_utils
from virttest import element_tree
from virttest.propcan import PropCanBase
from virttest.libvirt_xml import xcepts, base


def type_check(name, thing, expected):
    """
    Check that thing is expected subclass or instance, raise ValueError if not
    """
    is_a = type(thing)
    is_a_name = str(is_a)
    if not isinstance(expected, list):
        expected = [expected]
    for e in expected:
        try:
            it_is = issubclass(thing, e)
        except TypeError:
            it_is = isinstance(thing, e)
        if it_is:
            return
    raise ValueError('%s value is not any of %s, it is a %s'
                     % (name, expected, is_a_name))


def add_to_slots(*args):
    """
    Return list of AccessorBase.__all_slots__ + args
    """
    for slot in args:
        type_check('slot name', slot, str)
    return AccessorBase.__all_slots__ + args


class AccessorBase(PropCanBase):

    """
    Base class for a callable operating on a LibvirtXMLBase subclass instance
    """

    # Gets AccessorGeneratorBase subclass's required_accessor_data_keys added
    __slots__ = ('operation', 'property_name', 'libvirtxml')

    def __init__(self, operation, property_name, libvirtxml, **dargs):
        """
        Initialize accessor to operate on lvxml with accessor_data for property

        :param operation: Debug String for 'Getter', 'Setter', or 'Delter'
        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param dargs: Necessary for subclasses to extend required parameters
        """
        type_check('Parameter property_name', property_name, str)
        type_check('Operation attribute', operation, str)
        type_check('__slots__ attribute', self.__all_slots__, [tuple, list])
        type_check('Parameter libvirtxml', libvirtxml, base.LibvirtXMLBase)

        super(AccessorBase, self).__init__()

        self.__dict_set__('operation', operation)
        self.__dict_set__('property_name', property_name)
        self.__dict_set__('libvirtxml', libvirtxml)

        for slot in self.__all_slots__:
            if slot in AccessorBase.__all_slots__:
                continue  # already checked these
            # Don't care about value type
            if slot not in dargs:
                raise ValueError('Required accessor generator parameter %s'
                                 % slot)
            self.__dict_set__(slot, dargs[slot])

    # Subclass expected to override this and specify parameters
    __call__ = NotImplementedError

    def __repr__(self):
        return ("%s's %s for %s with %s"
                % (self.libvirtxml.__class__.__name__, self.operation,
                   self.property_name, str(dict(self))))

    def xmltreefile(self):
        """
        Retrieve xmltreefile instance from libvirtxml instance
        """
        return self.libvirtxml.xmltreefile

    def element_by_parent(self, parent_xpath, tag_name, create=True):
        """
        Retrieve/create an element instance at parent_xpath/tag_name

        :param parent_xpath: xpath of parent element
        :param tag_name: name of element under parent to retrieve/create
        :param create: True to create new element if not exist
        :return: ElementTree.Element instance
        :raise: LibvirtXMLError: If element not exist & create=False
        """
        type_check('parent_xpath', parent_xpath, str)
        type_check('tag_name', tag_name, str)
        parent_element = self.xmltreefile().find(parent_xpath)
        if (parent_element == self.xmltreefile().getroot() and
                parent_element.tag == tag_name):
            return parent_element
        excpt_str = ('Exception thrown from %s for property "%s" while'
                     ' looking for element tag "%s", on parent at xpath'
                     ' "%s", in XML\n%s\n' % (self.operation,
                                              self.property_name, tag_name, parent_xpath,
                                              str(self.xmltreefile())))
        if parent_element is None:
            if create:
                # This will only work for simple XPath strings
                self.xmltreefile().create_by_xpath(parent_xpath)
                parent_element = self.xmltreefile().find(parent_xpath)
            # if create or not, raise if not exist
            if parent_element is None:
                raise xcepts.LibvirtXMLAccessorError(excpt_str)
        try:
            element = parent_element.find(tag_name)
        except Exception:
            logging.error(excpt_str)
            raise
        if element is None:
            if create:  # Create the element
                element = xml_utils.ElementTree.SubElement(parent_element,
                                                           tag_name)
            else:  # create is False
                raise xcepts.LibvirtXMLNotFoundError('Error in %s for property '
                                                     '"%s", element tag "%s" not '
                                                     'found on parent at xpath "%s"'
                                                     ' in XML\n%s\n'
                                                     % (self.operation,
                                                        self.property_name,
                                                        tag_name, parent_xpath,
                                                        str(self.xmltreefile())))
        return element


class ForbiddenBase(AccessorBase):

    """
    Raise LibvirtXMLAccessorError when called w/ or w/o a value arg.
    """

    __slots__ = []

    def __call__(self, value=None):
        if value:
            raise xcepts.LibvirtXMLForbiddenError("%s %s to '%s' on %s "
                                                  "forbidden"
                                                  % (self.operation,
                                                     self.property_name,
                                                     str(value),
                                                     str(self)))
        else:
            raise xcepts.LibvirtXMLForbiddenError("%s %s on %s "
                                                  "forbidden"
                                                  % (self.operation,
                                                     self.property_name,
                                                     str(self)))


class AccessorGeneratorBase(object):

    """
    Accessor method/class generator for specific property name
    """

    def __init__(self, property_name, libvirtxml, forbidden=None, **dargs):
        """
        Initialize accessor methods, marking operations in forbidden as such

        :param property_name: Name of the property
        :param libvirtxml: Instance reference to LibvirtXMLBase subclass
        :param forbidden: Optional string list of 'get', 'set', and/or 'del'
        :param dargs: Specific AccessorGeneratorBase subclass info.
        """
        if forbidden is None:
            forbidden = []
        type_check('forbidden', forbidden, list)
        self.forbidden = forbidden

        type_check('libvirtxml', libvirtxml, base.LibvirtXMLBase)
        self.libvirtxml = libvirtxml

        type_check('property_name', property_name, str)
        self.property_name = property_name

        self.dargs = dargs

        # Lookup all property names possibly needing accessors
        for operation in ('get', 'set', 'del'):
            self.set_if_not_defined(operation)

    def set_if_not_defined(self, operation):
        """
        Setup a callable instance for operation only if not already defined
        """
        # Don't overwrite methods in libvirtxml instance
        if not hasattr(self.libvirtxml, self.accessor_name(operation)):
            if operation not in self.forbidden:
                self.assign_callable(operation, self.make_callable(operation))
            else:  # operation is forbidden
                self.assign_callable(operation, self.make_forbidden(operation))

    def accessor_name(self, operation):
        """
        Return instance name for operation, defined by subclass (i.e. 'get_foo')
        """
        return "%s_%s" % (operation, self.property_name)

    @staticmethod
    def callable_name(operation):
        """
        Return class name for operation (i.e. 'Getter'), defined by subclass.
        """
        return operation.capitalize() + 'ter'

    def make_callable(self, operation):
        """
        Return an callable instance for operation
        """
        callable_class = getattr(self, self.callable_name(operation))
        return callable_class(
            self.callable_name(operation), self.property_name,
            self.libvirtxml, **self.dargs)

    def make_forbidden(self, operation):
        """
        Return a forbidden callable instance for operation
        """
        return ForbiddenBase(operation, self.property_name, self.libvirtxml)

    def assign_callable(self, operation, callable_inst):
        """
        Set reference on objectified libvirtxml instance to callable_inst
        """
        self.libvirtxml.__super_set__(self.accessor_name(operation),
                                      callable_inst)


# Implementation of specific accessor generator subclasses follows


class AllForbidden(AccessorGeneratorBase):

    """
    Class of forbidden accessor classes for those undefined on libvirtxml
    """

    def __init__(self, property_name, libvirtxml):
        """
        Create exception raising accessors for those undefined on libvirtxml

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        """
        super(AllForbidden, self).__init__(property_name=property_name,
                                           libvirtxml=libvirtxml,
                                           forbidden=['get', 'set', 'del'])


class XMLElementText(AccessorGeneratorBase):

    """
    Class of accessor classes operating on element.text
    """

    required_dargs = ('parent_xpath', 'tag_name')

    def __init__(self, property_name, libvirtxml, forbidden=None,
                 parent_xpath=None, tag_name=None):
        """
        Create undefined accessors on libvirt instance

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param forbidden: Optional list of 'get', 'set', 'del'
        :param parent_xpath: XPath string of parent element
        :param tag_name: element tag name to manipulate text attribute on.
        """
        super(XMLElementText, self).__init__(property_name, libvirtxml,
                                             forbidden,
                                             parent_xpath=parent_xpath,
                                             tag_name=tag_name)

    class Getter(AccessorBase):

        """
        Retrieve text on element
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name')

        def __call__(self):
            return self.element_by_parent(self.parent_xpath,
                                          self.tag_name, create=False).text

    class Setter(AccessorBase):

        """
        Set text to value on element
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name')

        def __call__(self, value):
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=True)
            element.text = str(value)
            self.xmltreefile().write()

    class Delter(AccessorBase):

        """
        Remove element and ignore if it doesn't exist (same as False)
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name')

        def __call__(self):
            try:
                element = self.element_by_parent(self.parent_xpath,
                                                 self.tag_name, create=False)
            except (xcepts.LibvirtXMLNotFoundError,  # element doesn't exist
                    xcepts.LibvirtXMLAccessorError):  # parent doesn't exist
                pass  # already gone
            else:
                parent = self.xmltreefile().find(self.parent_xpath)
                if parent is not None:
                    parent.remove(element)
                    self.xmltreefile().write()


class XMLElementInt(AccessorGeneratorBase):

    """
    Class of accessor classes operating on element.text as an integer
    """
    __radix2func_dict__ = {0: int,
                           2: bin,
                           8: oct,
                           10: int,
                           16: hex}

    required_dargs = ('parent_xpath', 'tag_name', 'radix')

    def __init__(self, property_name, libvirtxml, forbidden=None,
                 parent_xpath=None, tag_name=None, radix=10):
        """
        Create undefined accessors on libvirt instance

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param forbidden: Optional list of 'Getter', 'Setter', 'Delter'
        :param parent_xpath: XPath string of parent element
        :param tag_name: element tag name to manipulate text attribute on.
        """
        try:
            self.__radix2func_dict__[radix]
        except KeyError:
            raise xcepts.LibvirtXMLError("Param radix=%s for XMLElementInt "
                                         "is not accepted." % radix)
        super(XMLElementInt, self).__init__(property_name, libvirtxml,
                                            forbidden,
                                            parent_xpath=parent_xpath,
                                            tag_name=tag_name,
                                            radix=radix)

    class Getter(AccessorBase):

        """
        Retrieve text on element and convert to int
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name', 'radix')

        def __call__(self):
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=False)
            try:
                result = int(element.text, self.radix)
            except ValueError:
                raise xcepts.LibvirtXMLError("Value of %s in %s is %s,"
                                             "not a Integer." % (self.tag_name,
                                                                 self.parent_xpath, element.text))
            return result

    class Setter(AccessorBase):

        """
        Set text on element after converting to int then to str
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name', 'radix')

        def __call__(self, value):
            type_check(self.property_name + ' value', value, int)
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=True)
            convertFunc = XMLElementInt.__radix2func_dict__[self.radix]
            element.text = str(convertFunc(value))
            self.xmltreefile().write()

    Delter = XMLElementText.Delter


class XMLElementBool(AccessorGeneratorBase):

    """
    Class of accessor classes operating purely element existence
    """

    required_dargs = ('parent_xpath', 'tag_name')

    def __init__(self, property_name, libvirtxml, forbidden=None,
                 parent_xpath=None, tag_name=None):
        """
        Create undefined accessors on libvirt instance

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param forbidden: Optional list of 'get', 'set', 'del'
        :param parent_xpath: XPath string of parent element
        :param tag_name: element tag name to manipulate text attribute on.
        """
        super(XMLElementBool, self).__init__(property_name, libvirtxml,
                                             forbidden,
                                             parent_xpath=parent_xpath,
                                             tag_name=tag_name)

    class Getter(AccessorBase):

        """
        Retrieve text on element
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name')

        def __call__(self):
            try:
                # Throws exception if parent path or element not exist
                self.element_by_parent(self.parent_xpath, self.tag_name,
                                       create=False)
                return True
            except (xcepts.LibvirtXMLAccessorError,
                    xcepts.LibvirtXMLNotFoundError):
                return False

    class Setter(AccessorBase):

        """
        Create element when True, delete when false
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name')

        def __call__(self, value):
            if bool(value) is True:
                self.element_by_parent(self.parent_xpath, self.tag_name,
                                       create=True)
            else:
                delattr(self.libvirtxml, self.property_name)
            self.xmltreefile().write()

    Delter = XMLElementText.Delter


class XMLAttribute(AccessorGeneratorBase):

    """
    Class of accessor classes operating on an attribute of an element
    """

    def __init__(self, property_name, libvirtxml, forbidden=None,
                 parent_xpath=None, tag_name=None, attribute=None):
        """
        Create undefined accessors on libvirt instance

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param forbidden: Optional list of 'Getter', 'Setter', 'Delter'
        :param parent_xpath: XPath string of parent element
        :param tag_name: element tag name to manipulate text attribute on.
        :param attribute: Attribute name to manupulate
        """
        super(XMLAttribute, self).__init__(property_name, libvirtxml,
                                           forbidden, parent_xpath=parent_xpath,
                                           tag_name=tag_name, attribute=attribute)

    class Getter(AccessorBase):

        """
        Get attribute value
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name', 'attribute')

        def __call__(self):
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=False)
            value = element.get(self.attribute, None)
            if value is None:
                raise xcepts.LibvirtXMLNotFoundError("Attribute %s not found "
                                                     "on element %s"
                                                     % (self.attribute,
                                                        element.tag))
            return value

    class Setter(AccessorBase):

        """
        Set attribute value
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name', 'attribute')

        def __call__(self, value):
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=True)
            element.set(self.attribute, str(value))
            self.xmltreefile().write()

    class Delter(AccessorBase):

        """
        Remove attribute
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name', 'attribute')

        def __call__(self):
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=False)
            try:
                del element.attrib[self.attribute]
            except KeyError:
                pass  # already doesn't exist
            self.xmltreefile().write()


class XMLElementDict(AccessorGeneratorBase):

    """
    Class of accessor classes operating as a dictionary of attributes
    """

    def __init__(self, property_name, libvirtxml, forbidden=None,
                 parent_xpath=None, tag_name=None):
        """
        Create undefined accessors on libvirt instance

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param forbidden: Optional list of 'Getter', 'Setter', 'Delter'
        :param parent_xpath: XPath string of parent element
        :param tag_name: element tag name to manipulate text attribute on.
        """
        super(XMLElementDict, self).__init__(property_name, libvirtxml,
                                             forbidden,
                                             parent_xpath=parent_xpath,
                                             tag_name=tag_name)

    class Getter(AccessorBase):

        """
        Retrieve attributes on element
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name')

        def __call__(self):
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=False)
            return dict(list(element.items()))

    class Setter(AccessorBase):

        """
        Set attributes to value on element
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name')

        def __call__(self, value):
            type_check(self.property_name + ' value', value, dict)
            element = self.element_by_parent(self.parent_xpath,
                                             self.tag_name, create=True)
            for attr_key, attr_value in list(value.items()):
                element.set(str(attr_key), str(attr_value))
            self.xmltreefile().write()

    # Inheriting from XMLElementText not work right
    Delter = XMLElementText.Delter


class XMLElementNest(AccessorGeneratorBase):

    """
    Class of accessor classes operating on a LibvirtXMLBase subclass
    """

    required_dargs = ('parent_xpath', 'tag_name', 'subclass', 'subclass_dargs')

    def __init__(self, property_name, libvirtxml, forbidden=None,
                 parent_xpath=None, tag_name=None, subclass=None,
                 subclass_dargs=None):
        """
        Create undefined accessors on libvirt instance

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param forbidden: Optional list of 'Getter', 'Setter', 'Delter'
        :param parent_xpath: XPath string of parent element
        :param tag_name: element tag name to manipulate text attribute on.
        :param subclass: A LibvirtXMLBase subclass with root tag == tag_name
        :param subclass_dargs: dict. to pass as kw args to subclass.__init__

        N/B: Works ONLY if tag_name is unique within parent element
        """
        type_check('subclass', subclass, base.LibvirtXMLBase)
        type_check('subclass_dargs', subclass_dargs, dict)
        super(XMLElementNest, self).__init__(property_name, libvirtxml,
                                             forbidden,
                                             parent_xpath=parent_xpath,
                                             tag_name=tag_name,
                                             subclass=subclass,
                                             subclass_dargs=subclass_dargs)

    class Getter(AccessorBase):

        """
        Retrieve instance of subclass with it's xml set to rerooted xpath/tag
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name', 'subclass',
                                 'subclass_dargs')

        def __call__(self):
            xmltreefile = self.xmltreefile()
            # Don't re-invent XPath generation method/behavior
            nested_root_element = self.element_by_parent(self.parent_xpath,
                                                         self.tag_name,
                                                         create=False)
            nested_root_xpath = xmltreefile.get_xpath(nested_root_element)
            # Try to make XMLTreeFile copy, rooted at nested_root_xpath
            # with copies of any/all child elements also
            nested_xtf = xmltreefile.reroot(nested_root_xpath)
            # Create instance of subclass to assign nested_xtf onto
            nestedinst = self.subclass(**self.subclass_dargs)
            # nestedxml.xmltreefile.restore() will fail on nested_xtf.__del__
            # set from string not filename!
            nestedinst.set_xml(str(nested_xtf))
            return nestedinst

    class Setter(AccessorBase):

        """
        Set attributes to value on element
        """

        __slots__ = add_to_slots('parent_xpath', 'tag_name', 'subclass')

        def __call__(self, value):
            type_check('Instance of %s' % self.subclass.__name__,
                       value,
                       self.subclass)
            # Will overwrite if exists
            existing_element = self.element_by_parent(self.parent_xpath,
                                                      self.tag_name,
                                                      create=True)
            existing_parent = self.xmltreefile().get_parent(existing_element)
            self.xmltreefile().remove(existing_element)
            existing_parent.append(value.xmltreefile.getroot())
            self.xmltreefile().write()

    # Nothing fancy, just make sure that part of tree doesn't exist
    Delter = XMLElementText.Delter


class XMLElementList(AccessorGeneratorBase):

    """
    Class of accessor classes operating on a list of child elements

    Other generators here have a hard-time dealing with XML that has
    multiple child-elements with the same tag.  This class allows
    treating these structures as lists of arbitrary user-defined
    objects.  User-defined marshal functions are called to perform
    the conversion to/from the format described in __init__.
    """

    required_dargs = ('parent_xpath', 'tag_name', 'marshal_from', 'marshal_to')

    def __init__(self, property_name, libvirtxml, forbidden=None,
                 parent_xpath=None, marshal_from=None, marshal_to=None,
                 has_subclass=None):
        """
        Create undefined accessors on libvirt instance

        :param property_name: String name of property (for exception detail)
        :param libvirtxml: An instance of a LibvirtXMLBase subclass
        :param forbidden: Optional list of 'Getter', 'Setter', 'Delter'
        :param parent_xpath: XPath string of parent element
        :param marshal_from: Callable, passed the item, index, and
                              libvirtxml instance.  Must return tuple
                              of tag-name, an attribute-dict, and an optional
                              text or raise ValueError exception.
        :param marshal_to: Callable. Passed a the item tag, attribute-dict, index,
                            libvirtxml, and optional text instance.  Returns
                            item value accepted by marshal_from or None to skip
        """
        if not callable(marshal_from) or not callable(marshal_to):
            raise ValueError("Both marshal_from and marshal_to must be "
                             "callable")
        super(XMLElementList, self).__init__(property_name, libvirtxml,
                                             forbidden,
                                             parent_xpath=parent_xpath,
                                             marshal_from=marshal_from,
                                             marshal_to=marshal_to,
                                             has_subclass=has_subclass)

    class Getter(AccessorBase):

        """
        Retrieve list of values as returned by the marshal_to callable
        """

        __slots__ = add_to_slots('parent_xpath', 'marshal_to', 'has_subclass')

        def __call__(self):
            # Parent structure cannot be pre-determined as in other classes
            parent = self.xmltreefile().find(self.parent_xpath)
            if parent is None:
                # Used as "undefined" signal, raising exception may
                # not be appropriate when other accessors are used
                # to generate missing structure.
                return None
            result = []
            # Give user-defined marshal functions a way to act on
            # item order if needed, and/or help with error reporting.
            index = 0
            # user-defined marshal functions might want to use
            # index numbers to filter/skip certain elements
            # but also support specific item ordering.
            for child in parent.getchildren():
                # Call user-defined helper to translate Element
                # into simple pre-defined format.

                # To support converting xml elements directly to a list of
                # xml objects, first create xmltreefile for new object
                if self.has_subclass:
                    new_xmltreefile = xml_utils.XMLTreeFile(
                        element_tree.tostring(child))
                    item = self.marshal_to(child.tag, new_xmltreefile,
                                           index, self.libvirtxml)
                else:
                    try:
                        # To support an optional text parameter, compatible
                        # with no text parameter.
                        item = self.marshal_to(child.tag, dict(list(child.items())),
                                               index, self.libvirtxml, child.text)
                    except TypeError:
                        item = self.marshal_to(child.tag, dict(list(child.items())),
                                               index, self.libvirtxml)
                if item is not None:
                    result.append(item)
                # Always use absolute index (even if item was None)
                index += 1
            return result

    class Setter(AccessorBase):

        """
        Set child elements as returned by the marshal_to callable
        """

        __slots__ = add_to_slots('parent_xpath', 'marshal_from', 'has_subclass')

        def __call__(self, value):
            type_check('value', value, list)
            # Allow other classes to generate parent structure
            parent = self.xmltreefile().find(self.parent_xpath)
            if parent is None:
                raise xcepts.LibvirtXMLNotFoundError
            # Remove existing by calling accessor method, allowing
            # any "untouchable" or "filtered" elements (by marshal)
            # to be ignored and left as-is.
            delattr(self.libvirtxml, self.property_name)
            # Allow user-defined marshal function to determine
            # if item order is important.  Also give more meaningful
            # exception message below, if there is a problem.
            index = 0
            for item in value:
                try:
                    # Call user-defined conversion from simple
                    # format, back to Element instances.
                    element_tuple = self.marshal_from(item, index,
                                                      self.libvirtxml)
                except ValueError:
                    # Defined in marshal API, to help with error reporting
                    # and debugging with more rich message.
                    msg = ("Call to %s by set accessor method for property %s "
                           "with unsupported item type %s, at index %d, "
                           " with value %s." % (str(self.marshal_from),
                                                self.property_name,
                                                str(type(item)),
                                                index,
                                                str(item)))
                    raise xcepts.LibvirtXMLAccessorError(msg)

                # Handle xml sub-element items
                if self.has_subclass:
                    parent.append(element_tuple[1].xmltreefile.getroot())
                    self.xmltreefile().write()
                    continue

                # Handle element text values in element_tuple[1].
                text = None
                new_dict = element_tuple[1].copy()
                if 'text' in list(element_tuple[1].keys()):
                    del new_dict['text']

                # To support text in element, marshal_from may return an
                # additional text value, check the length of element tuple
                if len(element_tuple) == 3:
                    text = element_tuple[2]
                parent_element = xml_utils.ElementTree.SubElement(
                    parent,
                    element_tuple[0],
                    new_dict,
                    text)

                # To support child element contains text, create sub element
                # with text under new created parent element.
                if 'text' in list(element_tuple[1].keys()):
                    text_dict = element_tuple[1]['text']
                    attr_dict = {}
                    for text_key, text_val in list(text_dict.items()):
                        xml_utils.ElementTree.SubElement(
                            parent_element,
                            text_key,
                            attr_dict,
                            text_val)

                index += 1
            self.xmltreefile().write()

    class Delter(AccessorBase):

        """
        Remove ALL child elements for which marshal_to does NOT return None
        """

        __slots__ = add_to_slots('parent_xpath', 'marshal_to', 'has_subclass')

        def __call__(self):
            parent = self.xmltreefile().find(self.parent_xpath)
            if parent is None:
                raise xcepts.LibvirtXMLNotFoundError("Parent element %s not "
                                                     "found" % self.parent_xpath)
            # Don't delete while traversing list
            todel = []
            index = 0
            for child in parent.getchildren():
                # To support directly deleting xml elements xml objects,
                # first create xmltreefile for new object
                if self.has_subclass:
                    new_xmltreefile = xml_utils.XMLTreeFile(
                        element_tree.tostring(child))
                    item = self.marshal_to(child.tag, new_xmltreefile,
                                           index, self.libvirtxml)
                else:
                    try:
                        # To support an optional text parameter, compatible
                        # with no text parameter.
                        item = self.marshal_to(child.tag, dict(list(child.items())),
                                               index, self.libvirtxml, child.text)
                    except TypeError:
                        item = self.marshal_to(child.tag, dict(list(child.items())),
                                               index, self.libvirtxml)
                # Always use absolute index (even if item was None)
                index += 1
                # Account for case where child elements are mixed in
                # with other elements not supported by this class.
                # Also permits marshal functions to do element filtering
                # if the class should only address specificly attributed
                # elements.
                if item is not None:
                    todel.append(child)
            for child in todel:
                parent.remove(child)



import gtk


class TestForm(gtk.Window):

    def __init__(self):
        super(TestForm, self).__init__()

        self.set_title("Key test")
        self.set_size_request(200, 200)
        self.set_position(gtk.WIN_POS_CENTER)

        fixed = gtk.Fixed()

        entry = gtk.Entry()
        fixed.put(entry, 10, 10)

        entry.connect("key_press_event", self.on_key_press_event)

        self.connect("destroy", gtk.main_quit)
        self.add(fixed)
        self.show_all()

        # Clean the text file:
        input_file = open("/tmp/autotest-rv_input", "w")
        input_file.close()

    def on_key_press_event(self, widget, event):
        # Store caught keycodes into text file
        input_file = open("/tmp/autotest-rv_input", "a")
        input_file.write("{0} ".format(event.keyval))
        input_file.close()


if __name__ == "__main__":
    TestForm()
    gtk.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2020
# Author: Sebastian Mitterle <smitterl@redhat.com>

"""
Handle Linux kernel modules
"""

import os
import logging

from avocado.utils import process


class KernelModuleError(Exception):

    def __init__(self, handler, module_name, reason):
        self.handler = handler
        self.module_name = module_name
        self.reason = reason

    def __str__(self):
        return "Couldn't %s module %s: %s" % (self.handler, self.module_name,
                                              self.reason)


class KernelModuleUnloadError(KernelModuleError):

    def __init__(self, module_name, reason):
        super(KernelModuleUnloadError, self).__init__("unload", module_name,
                                                      reason)


class KernelModuleReloadError(KernelModuleError):

    def __init__(self, module_name, reason):
        super(KernelModuleReloadError, self).__init__("reload", module_name,
                                                      reason)


class KernelModuleRestoreError(KernelModuleError):

    def __init__(self, module_name, reason):
        super(KernelModuleRestoreError, self).__init__("restore", module_name,
                                                       reason)


def reload(module_name, force, params=""):
    """
    Convenience method that creates a KernelModuleHandler instance
    and reloads the module only if any action will is required.

    :param module_name: name of kernel module to be handled
    :param force: if to force load with params in any case, e.g. True
    :param params: parameters to load with, e.g. 'key1=param1 ...'
    :return: instance if module was loaded
    """

    if params != "" or force:
        handler = KernelModuleHandler(module_name)
        handler.reload_module(force, params)
        return handler
    return None


class KernelModuleHandler(object):
    """Class handling Linux kernel modules"""

    def __init__(self, module_name):
        """Create kernel module handler"""

        self._module_name = module_name
        self._module_path = os.path.join('/sys/module/',
                                         self._module_name.replace('-', '_'))
        self._module_params_path = os.path.join(self._module_path, 'parameters')
        self._module_holders_path = os.path.join(self._module_path, "holders")
        self._was_loaded = None
        self._config_backup = None
        self._backup_config()

    def unload_module(self):
        """
        Unload module and those modules that use it.

        If there are some modules using this module, they are unloaded first.
        """
        if os.path.exists(self._module_path):
            unload_cmd = 'rmmod ' + self._module_name
            logging.debug("Unloading module: %s", unload_cmd)
            status, output = process.getstatusoutput(unload_cmd)
            if status:
                raise KernelModuleUnloadError(self._module_name, output)

    def reload_module(self, force, params=""):
        """
        Reload module with given parameters.

        If force=False loading will be skipped when either the module is
        already loaded with the passed parameters or when the module has
        not been loaded at all.

        +----------------------+-+-+-+-+-+-+
        |**precondition**                  |
        +----------------------+-+-+-+-+-+-+
        |module loaded         |N|N|Y|Y|Y|Y|
        +----------------------+-+-+-+-+-+-+
        |params already loaded |*|*|Y|N|Y|N|
        +----------------------+-+-+-+-+-+-+
        |force load            |Y|N|Y|Y|N|N|
        +----------------------+-+-+-+-+-+-+
        |**result**                        |
        +----------------------+-+-+-+-+-+-+
        |issue reload          |Y|N|Y|Y|N|Y|
        +----------------------+-+-+-+-+-+-+

        :param force: if to force load with params in any case, e.g. True
        :param params: parameters to load with, e.g. 'key1=param1 ...'
        """

        current_config = self.current_config
        if not force:
            do_not_load = False
            if (current_config and
                    all(x in current_config.split() for x in params.split())):
                logging.debug("Not reloading module. Current module config"
                              " uration for %s already contains all reques"
                              " ted parameters. Requested: '%s'. Current:"
                              " '%s'. Use force=True to force loading.",
                              self._module_name, params, current_config)
                do_not_load = True
            elif not self._was_loaded:
                logging.debug("Module %s isn't loaded. Use force=True to force"
                              " loading.", self._module_name)
                do_not_load = True
            if do_not_load:
                return

        # TODO: Handle cases were module cannot be removed
        holders = self.module_holders
        for holder in holders:
            holder.unload_module()
        self.unload_module()
        reload_cmd = 'modprobe %s %s' % (self._module_name, params)
        logging.debug("Reloading module: %s", reload_cmd)
        status, output = process.getstatusoutput(reload_cmd.strip())
        if status:
            raise KernelModuleReloadError(self._module_name, output)
        for holder in holders:
            holder.restore()

    def restore(self):
        """
        Restore previous module state.

        The state will only be restored if the original state
        was altered.

        +-------------------+-+-+-+-+
        |**precondition**           |
        +-------------------+-+-+-+-+
        |module loaded      |Y|Y|N|N|
        +-------------------+-+-+-+-+
        |loaded with params |Y|N|Y|N|
        +-------------------+-+-+-+-+
        |**result**                 |
        +-------------------+-+-+-+-+
        |issue restore      |Y|N|Y|N|
        +-------------------+-+-+-+-+
        """

        if self.current_config != self._config_backup:
            # TODO: Handle cases were module cannot be removed
            holders = self.module_holders
            for holder in holders:
                holder.unload_module()
            self.unload_module()
            if self._was_loaded:
                restore_cmd = 'modprobe %s %s' % (self._module_name,
                                                  self._config_backup)
                logging.debug("Restoring module state: %s", restore_cmd)
                status, output = process.getstatusoutput(restore_cmd)
                if status:
                    raise KernelModuleRestoreError(self._module_name,
                                                   output)
            for holder in holders:
                holder.restore()

    def _backup_config(self):
        """
        Check if self.module_name is loaded and read config

        """
        config = self._get_serialized_config()
        if config is not None:
            self._config_backup = config
            self._was_loaded = True
        else:
            self._was_loaded = False
        logging.debug("Backed up %s module state (was_loaded, params)"
                      "=(%s, %s)", self._module_name, self._was_loaded,
                      self._config_backup)

    @property
    def was_loaded(self):
        """ Read-only property """

        return self._was_loaded

    @property
    def config_backup(self):
        """ Read-only property """

        return self._config_backup

    @property
    def current_config(self):
        """ Read-only property """

        return self._get_serialized_config()

    def _get_serialized_config(self):
        """
        Get current module parameters

        :return: String holding module config 'param1=value1 param2=value2 ...', None if
         module not loaded
        """

        if not os.path.exists(self._module_params_path):
            return None

        mod_params = {}
        params = os.listdir(self._module_params_path)
        for param in params:
            with open(os.path.join(self._module_params_path,
                                   param), 'r') as param_file:
                mod_params[param] = param_file.read().strip()
        return " ".join("%s=%s" % _ for _ in mod_params.items()) if mod_params else ""

    @property
    def module_holders(self):
        """Find out which modules use this module."""
        if os.path.exists(self._module_holders_path):
            module_used_by = os.listdir(self._module_holders_path)
            return [KernelModuleHandler(module) for module in module_used_by]
        return []



"""
Utilities to create logical volumes or take snapshots of existing ones.

:author: Plamen Dimitrov
:copyright: Intra2net AG 2012
:license: GPL v2

:param vg_name: Name of the volume group.
:param lv_name: Name of the logical volume.
:param lv_size: Size of the logical volume as string in the form "#G"
                (for example 30G).
:param lv_snapshot_name: Name of the snapshot with origin the logical
                         volume.
:param lv_snapshot_size: Size of the snapshot with origin the logical
                         volume also as "#G".
:param ramdisk_vg_size: Size of the ramdisk virtual group.
:param ramdisk_basedir: Base directory for the ramdisk sparse file.
:param ramdisk_sparse_filename: Name of the ramdisk sparse file.

Sample ramdisk params:

::

    ramdisk_vg_size = "40000"
    ramdisk_basedir = "/tmp"
    ramdisk_sparse_filename = "virtual_hdd"

Sample general params:

::

    vg_name='autotest_vg',
    lv_name='autotest_lv',
    lv_size='1G',
    lv_snapshot_name='autotest_sn',
    lv_snapshot_size='1G'

The ramdisk volume group size is in MB.
"""

import logging
import re
import os
import shutil
import time

from avocado.core import exceptions
from avocado.utils import process

from virttest import error_context


@error_context.context_aware
def vg_ramdisk(vg_name, ramdisk_vg_size,
               ramdisk_basedir, ramdisk_sparse_filename):
    """
    Create vg on top of ram memory to speed up lv performance.
    """
    error_context.context("Creating virtual group on top of ram memory",
                          logging.info)
    vg_size = ramdisk_vg_size
    vg_ramdisk_dir = os.path.join(ramdisk_basedir, vg_name)
    ramdisk_filename = os.path.join(vg_ramdisk_dir,
                                    ramdisk_sparse_filename)

    vg_ramdisk_cleanup(ramdisk_filename,
                       vg_ramdisk_dir, vg_name, "")
    result = ""
    if not os.path.exists(vg_ramdisk_dir):
        os.mkdir(vg_ramdisk_dir)
    try:
        logging.info("Mounting tmpfs")
        result = process.run("mount -t tmpfs tmpfs " + vg_ramdisk_dir)

        logging.info("Converting and copying /dev/zero")
        cmd = ("dd if=/dev/zero of=" + ramdisk_filename +
               " bs=1M count=1 seek=" + vg_size)
        result = process.run(cmd, verbose=True)

        logging.info("Finding free loop device")
        result = process.run("losetup --find", verbose=True)
    except process.CmdError as ex:
        logging.error(ex)
        vg_ramdisk_cleanup(ramdisk_filename,
                           vg_ramdisk_dir, vg_name, "")
        raise ex

    loop_device = result.stdout_text.rstrip()

    try:
        logging.info("Creating loop device")
        result = process.run("losetup " + loop_device + " " + ramdisk_filename)
        logging.info("Creating physical volume %s", loop_device)
        result = process.run("pvcreate " + loop_device)
        logging.info("Creating volume group %s", vg_name)
        result = process.run("vgcreate " + vg_name + " " + loop_device)
    except process.CmdError as ex:
        logging.error(ex)
        vg_ramdisk_cleanup(ramdisk_filename, vg_ramdisk_dir, vg_name,
                           loop_device)
        raise ex

    logging.info(result.stdout_text.rstrip())


def vg_ramdisk_cleanup(ramdisk_filename, vg_ramdisk_dir,
                       vg_name, loop_device):
    """
    Inline cleanup function in case of test error.
    """
    result = process.run("vgremove " + vg_name, ignore_status=True)
    if result.exit_status == 0:
        logging.info(result.stdout_text.rstrip())
    else:
        logging.debug("%s -> %s", result.command, result.stderr_text)

    result = process.run("pvremove " + loop_device, ignore_status=True)
    if result.exit_status == 0:
        logging.info(result.stdout_text.rstrip())
    else:
        logging.debug("%s -> %s", result.command, result.stderr_text)

    for _ in range(10):
        time.sleep(0.1)
        result = process.run("losetup -d " + loop_device, ignore_status=True)
        if b"resource busy" not in result.stderr:
            if result.exit_status != 0:
                logging.debug("%s -> %s", result.command, result.stderr_text)
            else:
                logging.info("Loop device %s deleted", loop_device)
            break

    if os.path.exists(ramdisk_filename):
        os.unlink(ramdisk_filename)
        logging.info("Ramdisk filename %s deleted", ramdisk_filename)

    process.run("umount " + vg_ramdisk_dir, ignore_status=True)
    if result.exit_status == 0:
        if loop_device != "":
            logging.info("Loop device %s unmounted", loop_device)
    else:
        logging.debug("%s -> %s", result.command, result.stderr_text)

    if os.path.exists(vg_ramdisk_dir):
        try:
            shutil.rmtree(vg_ramdisk_dir)
            logging.info("Ramdisk directory %s deleted", vg_ramdisk_dir)
        except OSError:
            pass


def vg_check(vg_name):
    """
    Check whether provided volume group exists.
    """
    cmd = "vgdisplay " + vg_name
    try:
        process.run(cmd)
        logging.debug("Provided volume group exists: " + vg_name)
        return True
    except process.CmdError:
        return False


def vg_list():
    """
    List available volume groups.
    """
    cmd = "vgs --all"
    vgroups = {}
    result = process.run(cmd)

    lines = result.stdout_text.strip().splitlines()
    if len(lines) > 1:
        columns = lines[0].split()
        lines = lines[1:]
    else:
        return vgroups

    for line in lines:
        details = line.split()
        details_dict = {}
        index = 0
        for column in columns:
            if re.search("VG", column):
                vg_name = details[index]
            else:
                details_dict[column] = details[index]
            index += 1
        vgroups[vg_name] = details_dict
    return vgroups


@error_context.context_aware
def vg_create(vg_name, pv_list):
    """
    Create a volume group by using the block special devices
    """
    error_context.context(
        "Creating volume group '%s' by using '%s'" %
        (vg_name, pv_list), logging.info)

    if vg_check(vg_name):
        raise exceptions.TestError("Volume group '%s' already exist" % vg_name)
    # Update cached state before create VG
    cmd = "vgscan --cache"
    result = process.run(cmd, ignore_status=True)
    cmd = "vgcreate %s %s" % (vg_name, pv_list)
    result = process.run(cmd)
    logging.info(result.stdout_text.rstrip())


@error_context.context_aware
def vg_remove(vg_name):
    """
    Remove a volume group.
    """
    error_context.context("Removing volume '%s'" % vg_name, logging.info)

    if not vg_check(vg_name):
        raise exceptions.TestError(
            "Volume group '%s' could not be found" % vg_name)
    cmd = "vgremove -f %s" % vg_name
    result = process.run(cmd)
    logging.info(result.stdout_text.rstrip())
    # Update cached state after remove VG
    cmd = "vgscan --cache"
    result = process.run(cmd, ignore_status=True)


def lv_check(vg_name, lv_name):
    """
    Check whether provided logical volume exists.
    """
    cmd = "lvdisplay"
    result = process.run(cmd, ignore_status=True)

    # unstable approach but currently works
    lvpattern = r"LV Path\s+/dev/" + vg_name + r"/" + lv_name + "\s+"
    match = re.search(lvpattern, result.stdout_text.rstrip())
    if match:
        logging.debug("Provided logical volume exists: /dev/" +
                      vg_name + "/" + lv_name)
        return True
    else:
        return False


@error_context.context_aware
def lv_remove(vg_name, lv_name):
    """
    Remove a logical volume.
    """
    error_context.context("Removing volume /dev/%s/%s" %
                          (vg_name, lv_name), logging.info)

    if not vg_check(vg_name):
        raise exceptions.TestError("Volume group could not be found")
    if not lv_check(vg_name, lv_name):
        raise exceptions.TestError("Logical volume could not be found")

    cmd = "lvremove -f " + vg_name + "/" + lv_name
    result = process.run(cmd)
    logging.info(result.stdout_text.rstrip())


@error_context.context_aware
def lv_create(vg_name, lv_name, lv_size, force_flag=True):
    """
    Create a logical volume in a volume group.

    The volume group must already exist.
    """
    error_context.context("Creating original lv to take a snapshot from",
                          logging.info)

    if not vg_check(vg_name):
        raise exceptions.TestError("Volume group could not be found")
    if lv_check(vg_name, lv_name) and not force_flag:
        raise exceptions.TestError("Logical volume already exists")
    elif lv_check(vg_name, lv_name) and force_flag:
        lv_remove(vg_name, lv_name)

    cmd = ("lvcreate --size " + lv_size + " --name " + lv_name + " " + vg_name)
    result = process.run(cmd)
    logging.info(result.stdout_text.rstrip())


@error_context.context_aware
def lv_take_snapshot(vg_name, lv_name,
                     lv_snapshot_name, lv_snapshot_size):
    """
    Take a snapshot of the original logical volume.
    """
    error_context.context("Taking snapshot from original logical volume",
                          logging.info)

    if not vg_check(vg_name):
        raise exceptions.TestError("Volume group could not be found")
    if lv_check(vg_name, lv_snapshot_name):
        raise exceptions.TestError("Snapshot already exists")
    if not lv_check(vg_name, lv_name):
        raise exceptions.TestError("Snapshot's origin could not be found")

    cmd = ("lvcreate --size " + lv_snapshot_size + " --snapshot " +
           " --name " + lv_snapshot_name + " /dev/" + vg_name + "/" + lv_name)
    try:
        result = process.run(cmd)
    except process.CmdError as ex:
        if ('Logical volume "%s" already exists in volume group "%s"' %
            (lv_snapshot_name, vg_name) in ex.result.stderr_text and
            re.search(re.escape(lv_snapshot_name + " [active]"),
                      process.run("lvdisplay").stdout_text)):
            # the above conditions detect if merge of snapshot was postponed
            logging.warning(("Logical volume %s is still active! " +
                             "Attempting to deactivate..."), lv_name)
            lv_reactivate(vg_name, lv_name)
            result = process.run(cmd)
        else:
            raise ex
    logging.info(result.stdout_text.rstrip())


@error_context.context_aware
def lv_revert(vg_name, lv_name, lv_snapshot_name):
    """
    Revert the origin to a snapshot.
    """
    error_context.context("Reverting original logical volume to snapshot",
                          logging.info)
    try:
        if not vg_check(vg_name):
            raise exceptions.TestError("Volume group could not be found")
        if not lv_check(vg_name, lv_snapshot_name):
            raise exceptions.TestError("Snapshot could not be found")
        if (not lv_check(vg_name, lv_snapshot_name) and
                not lv_check(vg_name, lv_name)):
            raise exceptions.TestError(
                "Snapshot and its origin could not be found")
        if (lv_check(vg_name, lv_snapshot_name) and
                not lv_check(vg_name, lv_name)):
            raise exceptions.TestError("Snapshot origin could not be found")

        cmd = ("lvconvert --merge /dev/%s/%s" % (vg_name, lv_snapshot_name))
        result = process.run(cmd)
        if ("Merging of snapshot %s will start next activation." %
                lv_snapshot_name) in result.stdout_text:
            raise exceptions.TestError("The logical volume %s is still "
                                       "active" % lv_name)
        result = result.stdout_text.rstrip()

    except exceptions.TestError as ex:
        # detect if merge of snapshot was postponed
        # and attempt to reactivate the volume.
        ex_str = str(ex)
        if (('Snapshot could not be found' in ex_str and
             re.search(re.escape(lv_snapshot_name + " [active]"),
                       process.run("lvdisplay").stdout_text)) or
                ("The logical volume %s is still active" % lv_name) in ex_str):
            logging.warning(("Logical volume %s is still active! " +
                             "Attempting to deactivate..."), lv_name)
            lv_reactivate(vg_name, lv_name)
            result = "Continuing after reactivation"
        elif 'Snapshot could not be found' in ex_str:
            logging.error(ex)
            result = "Could not revert to snapshot"
        else:
            raise ex
    logging.info(result)


@error_context.context_aware
def lv_revert_with_snapshot(vg_name, lv_name,
                            lv_snapshot_name, lv_snapshot_size):
    """
    Perform logical volume merge with snapshot and take a new snapshot.

    """
    error_context.context("Reverting to snapshot and taking a new one",
                          logging.info)

    lv_revert(vg_name, lv_name, lv_snapshot_name)
    lv_take_snapshot(vg_name, lv_name, lv_snapshot_name, lv_snapshot_size)


@error_context.context_aware
def lv_reactivate(vg_name, lv_name, timeout=10):
    """
    In case of unclean shutdowns some of the lvs is still active and merging
    is postponed. Use this function to attempt to deactivate and reactivate
    all of them to cause the merge to happen.
    """
    try:
        process.run("lvchange -an /dev/%s/%s" % (vg_name, lv_name))
        time.sleep(timeout)
        process.run("lvchange -ay /dev/%s/%s" % (vg_name, lv_name))
        time.sleep(timeout)
    except process.CmdError:
        logging.error(("Failed to reactivate %s - please, " +
                       "nuke the process that uses it first."), lv_name)
        raise exceptions.TestError(
            "The logical volume %s is still active" % lv_name)


def gen_lvmap_mpath(session=None):
    """
    Return the name mapping between logical volume and device
    mapper multipathed device.
    :param session: When session is given, it means VM session
                    Otherwise default value means it work for host
    :return: None or a dict like:
            {"lv_name": dm_name}
    """
    names_map = None
    cmd = ("lvdisplay|awk '/LV Name/{n=$3} "
           "/Block device/{d=$3; sub(\".*:\", \"dm-\", d); print n,d;}'")

    if session:
        output = session.cmd_output(cmd).strip()
    else:
        output = process.run(cmd, shell=True).stdout_text
    lines = output.splitlines()
    if len(lines) >= 1:
        names_map = {}
        for line in lines:
            values = line.split()
            lv_name, mpath_name = values[0], values[1]
            names_map.update({lv_name: mpath_name})
    return names_map


def get_vg_mapped_blk_target(vg_name, session=None):
    """
    Get vg mapped blk target with pvs Command
    :param vg_name: Volume group Name
    :param session: When session is given, it means VM session
                    Otherwise default value means it work for host
    :return: The mapped blk target or None
    """
    cmd = "pvs | grep %s" % vg_name
    blk_target = None
    if session:
        output = session.cmd_output(cmd).strip()
    else:
        output = process.run(cmd, shell=True).stdout_text
    if output:
        # PV normally present as /dev/sda1
        pv_name = output.split()[0]
        blk_target = re.findall(r"[a-z]+", pv_name.split("/dev/")[1])[0]
    return blk_target



import re
import logging

try:
    import SocketServer as socketserver
except ImportError:
    import socketserver


SYSLOG_PORT = 514
DEFAULT_FORMAT = '[AutotestSyslog (%s.%s)] %s'


def set_default_format(message_format):
    '''
    Changes the default message format

    :type message_format: string
    :param message_format: a message format string with 3 placeholders:
                           facility, priority and message.
    '''
    global DEFAULT_FORMAT
    DEFAULT_FORMAT = message_format


def get_default_format():
    '''
    Returns the current default message format
    '''
    return DEFAULT_FORMAT


class RequestHandler(socketserver.BaseRequestHandler):

    '''
    A request handler that relays all received messages as DEBUG
    '''

    RECORD_RE = re.compile('\<(\d+)\>(.*)')

    (LOG_EMERG,
     LOG_ALERT,
     LOG_CRIT,
     LOG_ERR,
     LOG_WARNING,
     LOG_NOTICE,
     LOG_INFO,
     LOG_DEBUG) = list(range(8))

    (LOG_KERN,
     LOG_USER,
     LOG_MAIL,
     LOG_DAEMON,
     LOG_AUTH,
     LOG_SYSLOG,
     LOG_LPR,
     LOG_NEWS,
     LOG_UUCP,
     LOG_CRON,
     LOG_AUTHPRIV,
     LOG_FTP) = list(range(12))

    (LOG_LOCAL0,
     LOG_LOCAL1,
     LOG_LOCAL2,
     LOG_LOCAL3,
     LOG_LOCAL4,
     LOG_LOCAL5,
     LOG_LOCAL6,
     LOG_LOCAL7) = list(range(16, 24))

    PRIORITY_NAMES = {
        LOG_ALERT: "alert",
        LOG_CRIT: "critical",
        LOG_DEBUG: "debug",
        LOG_EMERG: "emerg",
        LOG_ERR: "err",
        LOG_INFO: "info",
        LOG_NOTICE: "notice",
        LOG_WARNING: "warning"
    }

    FACILITY_NAMES = {
        LOG_AUTH: "auth",
        LOG_AUTHPRIV: "authpriv",
        LOG_CRON: "cron",
        LOG_DAEMON: "daemon",
        LOG_FTP: "ftp",
        LOG_KERN: "kern",
        LOG_LPR: "lpr",
        LOG_MAIL: "mail",
        LOG_NEWS: "news",
        LOG_AUTH: "security",
        LOG_SYSLOG: "syslog",
        LOG_USER: "user",
        LOG_UUCP: "uucp",
        LOG_LOCAL0: "local0",
        LOG_LOCAL1: "local1",
        LOG_LOCAL2: "local2",
        LOG_LOCAL3: "local3",
        LOG_LOCAL4: "local4",
        LOG_LOCAL5: "local5",
        LOG_LOCAL6: "local6",
        LOG_LOCAL7: "local7",
    }

    def decodeFacilityPriority(self, priority):
        '''
        Decode both the facility and priority embedded in a syslog message

        :type priority: integer
        :param priority: an integer with facility and priority encoded
        :return: a tuple with two strings
        '''
        f = priority >> 3
        p = priority & 7
        return (self.FACILITY_NAMES.get(f, 'unknown'),
                self.PRIORITY_NAMES.get(p, 'unknown'))

    def log(self, data, message_format=None):
        '''
        Logs the received message as a DEBUG message
        '''
        match = self.RECORD_RE.match(data)
        if match:
            if message_format is None:
                message_format = get_default_format()
            pri = int(match.groups()[0])
            msg = match.groups()[1]
            (facility_name, priority_name) = self.decodeFacilityPriority(pri)
            logging.debug(message_format, facility_name, priority_name, msg)


class RequestHandlerTcp(RequestHandler):

    def handle(self):
        '''
        Handles a single request
        '''
        data = self.request.recv(4096)
        self.log(data)


class RequestHandlerUdp(RequestHandler):

    def handle(self):
        '''
        Handles a single request
        '''
        data = self.request[0]
        self.log(data)


class SysLogServerUdp(socketserver.UDPServer):

    def __init__(self, address):
        socketserver.UDPServer.__init__(self, address, RequestHandlerUdp)


class SysLogServerTcp(socketserver.TCPServer):

    def __init__(self, address):
        socketserver.TCPServer.__init__(self, address, RequestHandlerTcp)


def syslog_server(address='', port=SYSLOG_PORT,
                  tcp=True, terminate_callable=None):
    if tcp:
        klass = SysLogServerTcp
    else:
        klass = SysLogServerUdp
    syslog = klass((address, port))

    while True:
        if terminate_callable is not None:
            terminate = terminate_callable()
        else:
            terminate = False

        if not terminate:
            syslog.handle_request()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    syslog_server()



"""
smartcard device support class(es)

http://libvirt.org/formatdomain.html#elementsSmartcard
"""

from virttest.libvirt_xml import accessors
from virttest.libvirt_xml.devices import base


class Smartcard(base.TypedDeviceBase):
    __slots__ = ('smartcard_type', 'smartcard_mode', 'source', 'source_mode',
                 'source_host', 'source_service', 'protocol', 'protocol_type',
                 'address', 'address_type', 'address_controller', 'address_slot')

    def __init__(self, type_name='spicevmc', virsh_instance=base.base.virsh):
        accessors.XMLAttribute('smartcard_type', self,
                               parent_xpath='/',
                               tag_name='smartcard',
                               attribute='type')
        accessors.XMLAttribute('smartcard_mode', self,
                               parent_xpath='/',
                               tag_name='smartcard',
                               attribute='mode')
        accessors.XMLElementDict('source', self,
                                 parent_xpath='/',
                                 tag_name='source')
        accessors.XMLAttribute('source_mode', self,
                               parent_xpath='/',
                               tag_name='source',
                               attribute='mode')
        accessors.XMLAttribute('source_host', self,
                               parent_xpath='/',
                               tag_name='source',
                               attribute='host')
        accessors.XMLAttribute('source_service', self,
                               parent_xpath='/',
                               tag_name='source',
                               attribute='service')
        accessors.XMLElementDict('protocol', self,
                                 parent_xpath='/',
                                 tag_name='protocol')
        accessors.XMLAttribute('protocol_type', self,
                               parent_xpath='/',
                               tag_name='protocol',
                               attribute='type')
        accessors.XMLElementDict('address', self,
                                 parent_xpath='/',
                                 tag_name='address')
        accessors.XMLAttribute('address_type', self,
                               parent_xpath='/',
                               tag_name='address',
                               attribute='type')
        accessors.XMLAttribute('address_controller', self,
                               parent_xpath='/',
                               tag_name='address',
                               attribute='controller')
        accessors.XMLAttribute('address_slot', self,
                               parent_xpath='/',
                               tag_name='address',
                               attribute='slot')
        super(Smartcard, self).__init__(device_tag='smartcard',
                                        type_name=type_name,
                                        virsh_instance=virsh_instance)



'''
Created on Dec 11, 2013

:author: jzupka, astepano
:contact: Andrei Stepanov <astepano@redhat.com>
'''
import copy


class MessengerError(Exception):

    """
    Represented error in messanger.
    """

    def __init__(self, msg):
        super(MessengerError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return "Messenger ERROR %s" % (self.msg)


class CommanderError(MessengerError):

    """
    Represent error in Commnader
    """

    def __init__(self, msg):
        super(CommanderError, self).__init__(msg)

    def __str__(self):
        return "Commander ERROR %s" % (self.msg)


class CmdTraceBack(Exception):

    """
    Represent back-trace used for error tracing on remote side.
    """

    def __init__(self, msg):
        super(CmdTraceBack, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return "Cmd ERROR %s" % (self.msg)


class CmdMessage(object):

    """
    Base cmd message class
    """
    __slots__ = ["cmd_id"]

    def __init__(self, cmd_id):
        self.cmd_id = cmd_id

    def __getstate__(self):
        return (self.cmd_id)

    def __setstate__(self, state):
        self.cmd_id = state[0]

    def isCmdMsg(self):
        return self.cmd_id is not None

    def __eq__(self, other):
        return self.cmd_id == other.cmd_id


class StdStream(CmdMessage):

    """
    Represent message string data from remote client
    """
    __slots__ = ["msg"]

    def __init__(self, msg, cmd_id=None):
        super(StdStream, self).__init__(cmd_id)
        self.msg = msg

    def __str__(self):
        return (self.msg)

    def __getstate__(self):
        return (self.cmd_id, self.msg)

    def __setstate__(self, state):
        self.cmd_id = state[0]
        self.msg = state[1]


class StdOut(StdStream):

    """
    Represent message from stdout string data from remote client
    """
    __slots__ = ["cmd_id", "msg"]

    def __init__(self, msg, cmd_id=None):
        super(StdOut, self).__init__(msg, cmd_id)

    def __getstate__(self):
        return (self.cmd_id, self.msg)

    def __setstate__(self, state):
        self.cmd_id = state[0]
        self.msg = state[1]


class StdErr(StdStream):

    """
    Represent message from stderr string data from remote client
    """
    __slots__ = ["cmd_id", "msg"]

    def __init__(self, msg, cmd_id=None):
        super(StdErr, self).__init__(msg, cmd_id)

    def __getstate__(self):
        return (self.cmd_id, self.msg)

    def __setstate__(self, state):
        self.cmd_id = state[0]
        self.msg = state[1]


class CmdQuery(object):
    """Command-msg-request from VM to avocado-vt test.
    """
    def __init__(self, *args, **kargs):
        """
        Command for asking from VM to avocado-vt.

        :param args: Something pickable. Is irrelevant for messenger.
        :param kargs: Something pickable. Is irrelevant for messenger.
        """
        self.args = copy.deepcopy(args)
        self.kargs = copy.deepcopy(kargs)


class CmdRespond(object):
    """Command-msg-answer from avocado-test to VM.
    """
    def __init__(self, respond):
        """
        Command for answering avocado-vt to VM.

        :param respond: Something pickable. Is irrelevant for messenger.
        """
        self.respond = respond  # Must be pickable.


class BaseCmd(CmdMessage):

    """
    Class used for moveing information about commands between master and slave.
    """
    __slots__ = ["func", "args", "kargs", "results", "_async", "_finished",
                 "nh_stdin", "nh_stdout", "nh_stderr", "cmd_hash"]

    single_cmd_id = 0

    def __init__(self, func_cmd, *args, **kargs):
        self.cmd_id = BaseCmd.single_cmd_id
        BaseCmd.single_cmd_id += 1
        super(BaseCmd, self).__init__(self.cmd_id)

        self.func = func_cmd
        self.args = copy.deepcopy(args)
        self.kargs = copy.deepcopy(kargs)
        self.results = None
        self._async = False
        self._finished = False
        self.nh_stdin = None
        self.nh_stdout = None
        self.nh_stderr = None
        self.cmd_hash = None

    def __getstate__(self):
        return (self.cmd_id, self.func, self.args, self.kargs, self.results,
                self._async, self._finished, self.nh_stdin, self.nh_stdout,
                self.nh_stderr, self.cmd_hash)

    def __setstate__(self, state):
        self.cmd_id = state[0]
        self.func = state[1]
        self.args = state[2]
        self.kargs = state[3]
        self.results = state[4]
        self._async = state[5]
        self._finished = state[6]
        self.nh_stdin = state[7]
        self.nh_stdout = state[8]
        self.nh_stderr = state[9]
        self.cmd_hash = state[10]

    def __str__(self):
        str_args = []
        for a in self.args:  # Format str value in args to "val"
            if type(a) is str:
                str_args.append("\"%s\"" % a)
            else:
                str_args.append(a)

        str_kargs = {}
        for key, val in self.kargs:   # Format str value in kargs to "val"
            if type(val) is str:
                str_kargs[key] = "\"%s\"" % val
            else:
                str_kargs[key] = val

        return ("base_cmd: %s(%s)" % (".".join(self.func),
                                      ", ".join(str_args) +
                                      ",".join(list(str_kargs.items()))))

    def is_async(self):
        """
        :return: True if command is async else False
        """
        return self._async

    def is_finished(self):
        """
        :return: True if command is finished else False
        """
        return self._finished

    def update(self, basecmd):
        """
        Sync local class with class moved over the messanger.

        :param basecmd: basecmd from which should be sync data to this instance
        :type basecmd: BaseCmd
        """
        self.results = basecmd.results
        self._finished = basecmd._finished
        self._async = basecmd._async

    def update_cmd_hash(self, basecmd):
        if basecmd.cmd_hash is not None:
            self.cmd_hash = basecmd.cmd_hash
